#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Union, Iterable, Sequence, Optional, TYPE_CHECKING
import abc
import enum
import math

from ezdxf.lldxf import const
from ezdxf.lldxf.tags import Tags, group_tags
from ezdxf.math import (
    Vec2,
    Vec3,
    UVec,
    OCS,
    bulge_to_arc,
    ConstructionEllipse,
    ConstructionArc,
    BSpline,
    NonUniformScalingError,
    open_uniform_knot_vector,
    global_bspline_interpolation,
    arc_angle_span_deg,
    angle_to_param,
    param_to_angle,
)
from ezdxf.math.transformtools import OCSTransform

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = [
    "BoundaryPaths",
    "PolylinePath",
    "EdgePath",
    "LineEdge",
    "ArcEdge",
    "EllipseEdge",
    "SplineEdge",
    "EdgeType",
    "BoundaryPathType",
    "AbstractEdge",
    "AbstractBoundaryPath",
]


@enum.unique
class BoundaryPathType(enum.IntEnum):
    POLYLINE = 1
    EDGE = 2


@enum.unique
class EdgeType(enum.IntEnum):
    LINE = 1
    ARC = 2
    ELLIPSE = 3
    SPLINE = 4


class AbstractBoundaryPath(abc.ABC):
    type: BoundaryPathType
    path_type_flags: int
    source_boundary_objects: list[str]

    @abc.abstractmethod
    def clear(self) -> None: ...

    @classmethod
    @abc.abstractmethod
    def load_tags(cls, tags: Tags) -> AbstractBoundaryPath: ...

    @abc.abstractmethod
    def export_dxf(self, tagwriter: AbstractTagWriter, dxftype: str) -> None: ...

    @abc.abstractmethod
    def transform(self, ocs: OCSTransform, elevation: float) -> None: ...

    @abc.abstractmethod
    def is_valid(self) -> bool: ...


class AbstractEdge(abc.ABC):
    type: EdgeType

    @property
    @abc.abstractmethod
    def start_point(self) -> Vec2: ...

    @property
    @abc.abstractmethod
    def end_point(self) -> Vec2: ...

    @abc.abstractmethod
    def export_dxf(self, tagwriter: AbstractTagWriter) -> None: ...

    @abc.abstractmethod
    def transform(self, ocs: OCSTransform, elevation: float) -> None: ...

    @abc.abstractmethod
    def is_valid(self) -> bool: ...

    @property
    def real_start_point(self) -> Vec2:
        return self.start_point

    @property
    def real_end_point(self) -> Vec2:
        return self.end_point


class BoundaryPaths:
    def __init__(self, paths: Optional[list[AbstractBoundaryPath]] = None):
        self.paths: list[AbstractBoundaryPath] = paths or []

    def __len__(self):
        return len(self.paths)

    def __getitem__(self, item):
        return self.paths[item]

    def __iter__(self):
        return iter(self.paths)

    def is_valid(self) -> bool:
        return all(p.is_valid() for p in self.paths)

    @classmethod
    def load_tags(cls, tags: Tags) -> BoundaryPaths:
        paths = []
        assert tags[0].code == 92
        grouped_path_tags = group_tags(tags, splitcode=92)
        for path_tags in grouped_path_tags:
            path_type_flags = path_tags[0].value
            is_polyline_path = bool(path_type_flags & 2)
            path: AbstractBoundaryPath = (
                PolylinePath.load_tags(path_tags)
                if is_polyline_path
                else EdgePath.load_tags(path_tags)
            )
            path.path_type_flags = path_type_flags
            paths.append(path)
        return cls(paths)

    @property
    def has_edge_paths(self) -> bool:
        return any(p.type == BoundaryPathType.EDGE for p in self.paths)

    def clear(self) -> None:
        """Remove all boundary paths."""
        self.paths = []

    def external_paths(self) -> Iterable[AbstractBoundaryPath]:
        """Iterable of external paths, could be empty."""
        for b in self.paths:
            if b.path_type_flags & const.BOUNDARY_PATH_EXTERNAL:
                yield b

    def outermost_paths(self) -> Iterable[AbstractBoundaryPath]:
        """Iterable of outermost paths, could be empty."""
        for b in self.paths:
            if b.path_type_flags & const.BOUNDARY_PATH_OUTERMOST:
                yield b

    def default_paths(self) -> Iterable[AbstractBoundaryPath]:
        """Iterable of default paths, could be empty."""
        not_default = const.BOUNDARY_PATH_OUTERMOST + const.BOUNDARY_PATH_EXTERNAL
        for b in self.paths:
            if bool(b.path_type_flags & not_default) is False:
                yield b

    def rendering_paths(
        self, hatch_style: int = const.HATCH_STYLE_NESTED
    ) -> Iterable[AbstractBoundaryPath]:
        """Iterable of paths to process for rendering, filters unused
        boundary paths according to the given hatch style:

        - NESTED: use all boundary paths
        - OUTERMOST: use EXTERNAL and OUTERMOST boundary paths
        - IGNORE: ignore all paths except EXTERNAL boundary paths

        Yields paths in order of EXTERNAL, OUTERMOST and DEFAULT.

        """

        def path_type_enum(flags) -> int:
            if flags & const.BOUNDARY_PATH_EXTERNAL:
                return 0
            elif flags & const.BOUNDARY_PATH_OUTERMOST:
                return 1
            return 2

        paths = sorted(
            (path_type_enum(p.path_type_flags), i, p) for i, p in enumerate(self.paths)
        )
        ignore = 1  # EXTERNAL only
        if hatch_style == const.HATCH_STYLE_NESTED:
            ignore = 3
        elif hatch_style == const.HATCH_STYLE_OUTERMOST:
            ignore = 2
        return (p for path_type, _, p in paths if path_type < ignore)

    def append(self, path: AbstractBoundaryPath) -> None:
        """Append a new boundary path.

        .. versionadded:: 1.4
        """
        if not isinstance(path, AbstractBoundaryPath):
            raise TypeError(f"invalid path type: {type(path)}")
        self.paths.append(path)

    def add_polyline_path(
        self,
        path_vertices: Iterable[tuple[float, ...]],
        is_closed: bool = True,
        flags: int = 1,
    ) -> PolylinePath:
        """Create and add a new :class:`PolylinePath` object.

        Args:
            path_vertices: iterable of polyline vertices as (x, y) or
                (x, y, bulge)-tuples.
            is_closed: 1 for a closed polyline else 0
            flags: default(0), external(1), derived(4), textbox(8) or outermost(16)

        """
        new_path = PolylinePath.from_vertices(path_vertices, is_closed, flags)
        self.append(new_path)
        return new_path

    def add_edge_path(self, flags: int = 1) -> EdgePath:
        """Create and add a new :class:`EdgePath` object.

        Args:
            flags: default(0), external(1), derived(4), textbox(8) or outermost(16)

        """
        new_path = EdgePath()
        new_path.path_type_flags = flags
        self.append(new_path)
        return new_path

    def export_dxf(self, tagwriter: AbstractTagWriter, dxftype: str) -> None:
        tagwriter.write_tag2(91, len(self.paths))
        for path in self.paths:
            path.export_dxf(tagwriter, dxftype)

    def transform(self, ocs: OCSTransform, elevation: float = 0) -> None:
        """Transform HATCH boundary paths.

        These paths are 2d elements, placed in the OCS of the HATCH.

        """
        if not ocs.scale_uniform:
            self.polyline_to_edge_paths(just_with_bulge=True)
            self.arc_edges_to_ellipse_edges()

        for path in self.paths:
            path.transform(ocs, elevation=elevation)

    def polyline_to_edge_paths(self, just_with_bulge=True) -> None:
        """Convert polyline paths including bulge values to line- and arc edges.

        Args:
            just_with_bulge: convert only polyline paths including bulge
                values if ``True``

        """

        def _edges(points) -> Iterable[Union[LineEdge, ArcEdge]]:
            prev_point: Vec3 | None = None
            prev_bulge: float = 0.0
            for x, y, bulge in points:
                point = Vec3(x, y)
                if prev_point is None:
                    prev_point = point
                    prev_bulge = bulge
                    continue

                if prev_bulge != 0.0:
                    arc = ArcEdge()
                    # bulge_to_arc returns always counter-clockwise oriented
                    # start- and end angles:
                    (
                        arc.center,
                        start_angle,
                        end_angle,
                        arc.radius,
                    ) = bulge_to_arc(prev_point, point, prev_bulge)
                    chk_point = arc.center + Vec2.from_angle(start_angle, arc.radius)
                    arc.ccw = chk_point.isclose(prev_point, abs_tol=1e-9)
                    arc.start_angle = math.degrees(start_angle) % 360.0
                    arc.end_angle = math.degrees(end_angle) % 360.0
                    if math.isclose(arc.start_angle, arc.end_angle) and math.isclose(
                        arc.start_angle, 0
                    ):
                        arc.end_angle = 360.0
                    yield arc
                else:
                    line = LineEdge()
                    line.start = (prev_point.x, prev_point.y)
                    line.end = (point.x, point.y)
                    yield line

                prev_point = point
                prev_bulge = bulge

        def to_edge_path(polyline_path) -> EdgePath:
            edge_path = EdgePath()
            vertices: list = list(polyline_path.vertices)
            if polyline_path.is_closed:
                vertices.append(vertices[0])
            edge_path.edges = list(_edges(vertices))
            return edge_path

        for path_index, path in enumerate(self.paths):
            if isinstance(path, PolylinePath):
                if just_with_bulge and not path.has_bulge():
                    continue
                self.paths[path_index] = to_edge_path(path)

    def edge_to_polyline_paths(self, distance: float, segments: int = 16):
        """Convert all edge paths to simple polyline paths without bulges.

        Args:
            distance: maximum distance from the center of the curve to the
                center of the line segment between two approximation points to
                determine if a segment should be subdivided.
            segments: minimum segment count per curve

        """
        converted = []
        for path in self.paths:
            if path.type == BoundaryPathType.EDGE:
                path = flatten_to_polyline_path(path, distance, segments)
            converted.append(path)
        self.paths = converted

    def arc_edges_to_ellipse_edges(self) -> None:
        """Convert all arc edges to ellipse edges."""

        def to_ellipse(arc: ArcEdge) -> EllipseEdge:
            ellipse = EllipseEdge()
            ellipse.center = arc.center
            ellipse.ratio = 1.0
            ellipse.major_axis = (arc.radius, 0.0)
            ellipse.start_angle = arc.start_angle
            ellipse.end_angle = arc.end_angle
            ellipse.ccw = arc.ccw
            return ellipse

        for path in self.paths:
            if isinstance(path, EdgePath):
                edges = path.edges
                for edge_index, edge in enumerate(edges):
                    if isinstance(edge, ArcEdge):
                        edges[edge_index] = to_ellipse(edge)

    def ellipse_edges_to_spline_edges(self, num: int = 32) -> None:
        """
        Convert all ellipse edges to spline edges (approximation).

        Args:
            num: count of control points for a **full** ellipse, partial
                ellipses have proportional fewer control points but at least 3.

        """

        def to_spline_edge(e: EllipseEdge) -> SplineEdge:
            # No OCS transformation needed, source ellipse and target spline
            # reside in the same OCS.
            ellipse = ConstructionEllipse(
                center=e.center,
                major_axis=e.major_axis,
                ratio=e.ratio,
                start_param=e.start_param,
                end_param=e.end_param,
            )
            count = max(int(float(num) * ellipse.param_span / math.tau), 3)
            tool = BSpline.ellipse_approximation(ellipse, count)
            spline = SplineEdge()
            spline.degree = tool.degree
            if not e.ccw:
                tool = tool.reverse()

            spline.control_points = Vec2.list(tool.control_points)
            spline.knot_values = tool.knots()  # type: ignore
            spline.weights = tool.weights()  # type: ignore
            return spline

        for path_index, path in enumerate(self.paths):
            if isinstance(path, EdgePath):
                edges = path.edges
                for edge_index, edge in enumerate(edges):
                    if isinstance(edge, EllipseEdge):
                        edges[edge_index] = to_spline_edge(edge)

    def spline_edges_to_line_edges(self, factor: int = 8) -> None:
        """Convert all spline edges to line edges (approximation).

        Args:
            factor: count of approximation segments = count of control
                points x factor

        """

        def to_line_edges(spline_edge: SplineEdge):
            weights = spline_edge.weights
            if len(spline_edge.control_points):
                bspline = BSpline(
                    control_points=spline_edge.control_points,
                    order=spline_edge.degree + 1,
                    knots=spline_edge.knot_values,
                    weights=weights if len(weights) else None,
                )
            elif len(spline_edge.fit_points):
                bspline = BSpline.from_fit_points(
                    spline_edge.fit_points, spline_edge.degree
                )
            else:
                raise const.DXFStructureError(
                    "SplineEdge() without control points or fit points."
                )
            segment_count = (max(len(bspline.control_points), 3) - 1) * factor
            vertices = list(bspline.approximate(segment_count))
            for v1, v2 in zip(vertices[:-1], vertices[1:]):
                edge = LineEdge()
                edge.start = v1.vec2
                edge.end = v2.vec2
                yield edge

        for path in self.paths:
            if isinstance(path, EdgePath):
                new_edges = []
                for edge in path.edges:
                    if isinstance(edge, SplineEdge):
                        new_edges.extend(to_line_edges(edge))
                    else:
                        new_edges.append(edge)
                path.edges = new_edges

    def ellipse_edges_to_line_edges(self, num: int = 64) -> None:
        """Convert all ellipse edges to line edges (approximation).

        Args:
            num: count of control points for a **full** ellipse, partial
                ellipses have proportional fewer control points but at least 3.

        """

        def to_line_edges(edge):
            # Start- and end params are always stored in counter clockwise order!
            ellipse = ConstructionEllipse(
                center=edge.center,
                major_axis=edge.major_axis,
                ratio=edge.ratio,
                start_param=edge.start_param,
                end_param=edge.end_param,
            )
            segment_count = max(int(float(num) * ellipse.param_span / math.tau), 3)
            params = ellipse.params(segment_count + 1)

            # Reverse path if necessary!
            if not edge.ccw:
                params = reversed(list(params))
            vertices = list(ellipse.vertices(params))
            for v1, v2 in zip(vertices[:-1], vertices[1:]):
                line = LineEdge()
                line.start = v1.vec2
                line.end = v2.vec2
                yield line

        for path in self.paths:
            if isinstance(path, EdgePath):
                new_edges = []
                for edge in path.edges:
                    if isinstance(edge, EllipseEdge):
                        new_edges.extend(to_line_edges(edge))
                    else:
                        new_edges.append(edge)
                path.edges = new_edges

    def all_to_spline_edges(self, num: int = 64) -> None:
        """Convert all bulge, arc and ellipse edges to spline edges
        (approximation).

        Args:
            num: count of control points for a **full** circle/ellipse, partial
                circles/ellipses have proportional fewer control points but at
                least 3.

        """
        self.polyline_to_edge_paths(just_with_bulge=True)
        self.arc_edges_to_ellipse_edges()
        self.ellipse_edges_to_spline_edges(num)

    def all_to_line_edges(self, num: int = 64, spline_factor: int = 8) -> None:
        """Convert all bulge, arc and ellipse edges to spline edges and
        approximate this splines by line edges.

        Args:
            num: count of control points for a **full** circle/ellipse, partial
                circles/ellipses have proportional fewer control points but at
                least 3.
            spline_factor: count of spline approximation segments = count of
                control points x spline_factor

        """
        self.polyline_to_edge_paths(just_with_bulge=True)
        self.arc_edges_to_ellipse_edges()
        self.ellipse_edges_to_line_edges(num)
        self.spline_edges_to_line_edges(spline_factor)

    def has_critical_elements(self) -> bool:
        """Returns ``True`` if any boundary path has bulge values or arc edges
        or ellipse edges.

        """
        for path in self.paths:
            if isinstance(path, PolylinePath):
                return path.has_bulge()
            elif isinstance(path, EdgePath):
                for edge in path.edges:
                    if edge.type in {EdgeType.ARC, EdgeType.ELLIPSE}:
                        return True
            else:
                raise TypeError(type(path))
        return False


def flatten_to_polyline_path(
    path: AbstractBoundaryPath, distance: float, segments: int = 16
) -> PolylinePath:
    import ezdxf.path  # avoid cyclic imports

    # keep path in original OCS!
    ez_path = ezdxf.path.from_hatch_boundary_path(path, ocs=OCS(), elevation=0)
    vertices = ((v.x, v.y) for v in ez_path.flattening(distance, segments))
    return PolylinePath.from_vertices(
        vertices,
        flags=path.path_type_flags,
    )


def pop_source_boundary_objects_tags(path_tags: Tags) -> list[str]:
    source_boundary_object_tags = []
    while len(path_tags):
        if path_tags[-1].code in (97, 330):
            last_tag = path_tags.pop()
            if last_tag.code == 330:
                source_boundary_object_tags.append(last_tag.value)
            else:  # code == 97
                # result list does not contain the length tag!
                source_boundary_object_tags.reverse()
                return source_boundary_object_tags
        else:
            break
            # No source boundary objects found - entity is not valid for AutoCAD
    return []


def export_source_boundary_objects(
    tagwriter: AbstractTagWriter, handles: Sequence[str]
):
    tagwriter.write_tag2(97, len(handles))
    for handle in handles:
        tagwriter.write_tag2(330, handle)


class PolylinePath(AbstractBoundaryPath):
    type = BoundaryPathType.POLYLINE

    def __init__(self) -> None:
        self.path_type_flags: int = const.BOUNDARY_PATH_POLYLINE
        self.is_closed = False
        # list of 2D coordinates with bulge values (x, y, bulge);
        # bulge default = 0.0
        self.vertices: list[tuple[float, float, float]] = []
        # MPOLYGON does not support source boundary objects, the MPOLYGON is
        # the source object!
        self.source_boundary_objects: list[str] = []  # (330, handle) tags

    def is_valid(self) -> bool:
        return True

    @classmethod
    def from_vertices(
        cls,
        path_vertices: Iterable[tuple[float, ...]],
        is_closed: bool = True,
        flags: int = 1,
    ) -> PolylinePath:
        """Create a new :class:`PolylinePath` object from vertices.

        Args:
            path_vertices: iterable of polyline vertices as (x, y) or
                (x, y, bulge)-tuples.
            is_closed: 1 for a closed polyline else 0
            flags: default(0), external(1), derived(4), textbox(8) or outermost(16)

        """
        new_path = PolylinePath()
        new_path.set_vertices(path_vertices, is_closed)
        new_path.path_type_flags = flags | const.BOUNDARY_PATH_POLYLINE
        return new_path

    @classmethod
    def load_tags(cls, tags: Tags) -> PolylinePath:
        path = PolylinePath()
        path.source_boundary_objects = pop_source_boundary_objects_tags(tags)
        for tag in tags:
            code, value = tag
            if code == 10:  # vertex coordinates
                # (x, y, bulge); bulge default = 0.0
                path.vertices.append((value[0], value[1], 0.0))
            elif code == 42:  # bulge value
                x, y, bulge = path.vertices.pop()
                # Last coordinates with new bulge value
                path.vertices.append((x, y, value))
            elif code == 72:
                pass  # ignore this value
            elif code == 73:
                path.is_closed = value
            elif code == 92:
                path.path_type_flags = value
            elif code == 93:  # number of polyline vertices
                pass  # ignore this value
        return path

    def set_vertices(
        self, vertices: Iterable[Sequence[float]], is_closed: bool = True
    ) -> None:
        """Set new `vertices` as new polyline path, a vertex has to be a
        (x, y) or a (x, y, bulge)-tuple.

        """
        new_vertices = []
        for vertex in vertices:
            if len(vertex) == 2:
                x, y = vertex
                bulge = 0.0
            elif len(vertex) == 3:
                x, y, bulge = vertex
            else:
                raise const.DXFValueError(
                    "Invalid vertex format, expected (x, y) or (x, y, bulge)"
                )
            new_vertices.append((x, y, bulge))
        self.vertices = new_vertices
        self.is_closed = is_closed

    def clear(self) -> None:
        """Removes all vertices and all handles to associated DXF objects
        (:attr:`source_boundary_objects`).
        """
        self.vertices = []
        self.is_closed = False
        self.source_boundary_objects = []

    def has_bulge(self) -> bool:
        return any(bulge for x, y, bulge in self.vertices)

    def export_dxf(self, tagwriter: AbstractTagWriter, dxftype: str) -> None:
        has_bulge = self.has_bulge()
        write_tag = tagwriter.write_tag2

        write_tag(92, int(self.path_type_flags))
        if dxftype == "HATCH":  # great design :<
            write_tag(72, int(has_bulge))
            write_tag(73, int(self.is_closed))
        elif dxftype == "MPOLYGON":
            write_tag(73, int(self.is_closed))
            write_tag(72, int(has_bulge))
        else:
            raise ValueError(f"unsupported DXF type {dxftype}")
        write_tag(93, len(self.vertices))
        for x, y, bulge in self.vertices:
            tagwriter.write_vertex(10, (float(x), float(y)))
            if has_bulge:
                write_tag(42, float(bulge))

        if dxftype == "HATCH":
            export_source_boundary_objects(tagwriter, self.source_boundary_objects)

    def transform(self, ocs: OCSTransform, elevation: float) -> None:
        """Transform polyline path."""
        has_non_uniform_scaling = not ocs.scale_uniform

        def _transform():
            for x, y, bulge in self.vertices:
                # PolylinePaths() with arcs should be converted to
                # EdgePath(in BoundaryPath.transform()).
                if bulge and has_non_uniform_scaling:
                    raise NonUniformScalingError(
                        "Polyline path with arcs does not support non-uniform "
                        "scaling"
                    )
                v = ocs.transform_vertex(Vec3(x, y, elevation))
                yield v.x, v.y, bulge

        if self.vertices:
            self.vertices = list(_transform())


class EdgePath(AbstractBoundaryPath):
    type = BoundaryPathType.EDGE

    def __init__(self) -> None:
        self.path_type_flags = const.BOUNDARY_PATH_DEFAULT
        self.edges: list[AbstractEdge] = []
        self.source_boundary_objects = []

    def __iter__(self):
        return iter(self.edges)

    def is_valid(self) -> bool:
        return all(e.is_valid() for e in self.edges)

    @classmethod
    def load_tags(cls, tags: Tags) -> EdgePath:
        edge_path = cls()
        edge_path.source_boundary_objects = pop_source_boundary_objects_tags(tags)
        for edge_tags in group_tags(tags, splitcode=72):
            if len(edge_tags) == 0:
                continue
            edge_type = edge_tags[0].value
            if 0 < edge_type < 5:
                edge_path.edges.append(EDGE_CLASSES[edge_type].load_tags(edge_tags[1:]))
        return edge_path

    def transform(self, ocs: OCSTransform, elevation: float) -> None:
        """Transform edge boundary paths."""
        for edge in self.edges:
            edge.transform(ocs, elevation=elevation)

    def add_line(self, start: UVec, end: UVec) -> LineEdge:
        """Add a :class:`LineEdge` from `start` to `end`.

        Args:
            start: start point of line, (x, y)-tuple
            end: end point of line, (x, y)-tuple

        """
        line = LineEdge()
        line.start = Vec2(start)
        line.end = Vec2(end)
        self.edges.append(line)
        return line

    def add_arc(
        self,
        center: UVec,
        radius: float = 1.0,
        start_angle: float = 0.0,
        end_angle: float = 360.0,
        ccw: bool = True,
    ) -> ArcEdge:
        """Add an :class:`ArcEdge`.

        **Adding Clockwise Oriented Arcs:**

        Clockwise oriented :class:`ArcEdge` objects are sometimes necessary to
        build closed loops, but the :class:`ArcEdge` objects are always
        represented in counter-clockwise orientation.
        To add a clockwise oriented :class:`ArcEdge` you have to swap the
        start- and end angle and set the `ccw` flag to ``False``,
        e.g. to add a clockwise oriented :class:`ArcEdge` from 180 to 90 degree,
        add the :class:`ArcEdge` in counter-clockwise orientation with swapped
        angles::

            edge_path.add_arc(center, radius, start_angle=90, end_angle=180, ccw=False)

        Args:
            center: center point of arc, (x, y)-tuple
            radius: radius of circle
            start_angle: start angle of arc in degrees (`end_angle` for a
                clockwise oriented arc)
            end_angle: end angle of arc in degrees (`start_angle` for a
                clockwise oriented arc)
            ccw: ``True`` for counter-clockwise ``False`` for
                clockwise orientation

        """
        arc = ArcEdge()
        arc.center = Vec2(center)
        arc.radius = radius
        # Start- and end angles always for counter-clockwise oriented arcs!
        arc.start_angle = start_angle
        arc.end_angle = end_angle
        # Flag to export the counter-clockwise oriented arc in
        # correct clockwise orientation:
        arc.ccw = bool(ccw)
        self.edges.append(arc)
        return arc

    def add_ellipse(
        self,
        center: UVec,
        major_axis: UVec = (1.0, 0.0),
        ratio: float = 1.0,
        start_angle: float = 0.0,
        end_angle: float = 360.0,
        ccw: bool = True,
    ) -> EllipseEdge:
        """Add an :class:`EllipseEdge`.

        **Adding Clockwise Oriented Ellipses:**

        Clockwise oriented :class:`EllipseEdge` objects are sometimes necessary
        to build closed loops, but the :class:`EllipseEdge` objects are always
        represented in counter-clockwise orientation.
        To add a clockwise oriented :class:`EllipseEdge` you have to swap the
        start- and end angle and set the `ccw` flag to ``False``,
        e.g. to add a clockwise oriented :class:`EllipseEdge` from 180 to 90
        degree, add the :class:`EllipseEdge` in counter-clockwise orientation
        with swapped angles::

            edge_path.add_ellipse(center, major_axis, ratio, start_angle=90, end_angle=180, ccw=False)

        Args:
            center: center point of ellipse, (x, y)-tuple
            major_axis: vector of major axis as (x, y)-tuple
            ratio: ratio of minor axis to major axis as float
            start_angle: start angle of ellipse in degrees (`end_angle` for a
                clockwise oriented ellipse)
            end_angle: end angle of ellipse in degrees (`start_angle` for a
                clockwise oriented ellipse)
            ccw: ``True`` for counter-clockwise ``False`` for
                clockwise orientation

        """

        if ratio > 1.0:
            raise const.DXFValueError("argument 'ratio' has to be <= 1.0")
        ellipse = EllipseEdge()
        ellipse.center = Vec2(center)
        ellipse.major_axis = Vec2(major_axis)
        ellipse.ratio = ratio
        # Start- and end angles are always stored in counter-clockwise
        # orientation!
        ellipse.start_angle = start_angle
        ellipse.end_angle = end_angle
        # Flag to export the counter-clockwise oriented ellipse in
        # correct clockwise orientation:
        ellipse.ccw = bool(ccw)
        self.edges.append(ellipse)
        return ellipse

    def add_spline(
        self,
        fit_points: Optional[Iterable[UVec]] = None,
        control_points: Optional[Iterable[UVec]] = None,
        knot_values: Optional[Iterable[float]] = None,
        weights: Optional[Iterable[float]] = None,
        degree: int = 3,
        periodic: int = 0,
        start_tangent: Optional[UVec] = None,
        end_tangent: Optional[UVec] = None,
    ) -> SplineEdge:
        """Add a :class:`SplineEdge`.

        Args:
            fit_points: points through which the spline must go, at least 3 fit
                points are required. list of (x, y)-tuples
            control_points: affects the shape of the spline, mandatory and
                AutoCAD crashes on invalid data. list of (x, y)-tuples
            knot_values: (knot vector) mandatory and AutoCAD crashes on invalid
                data. list of floats; `ezdxf` provides two tool functions to
                calculate valid knot values: :func:`ezdxf.math.uniform_knot_vector`,
                :func:`ezdxf.math.open_uniform_knot_vector` (default if ``None``)
            weights: weight of control point, not mandatory, list of floats.
            degree: degree of spline (int)
            periodic: 1 for periodic spline, 0 for none periodic spline
            start_tangent: start_tangent as 2d vector, optional
            end_tangent: end_tangent as 2d vector, optional

        .. warning::

            Unlike for the spline entity AutoCAD does not calculate the
            necessary `knot_values` for the spline edge itself. On the contrary,
            if the `knot_values` in the spline edge are missing or invalid
            AutoCAD **crashes**.

        """
        spline = SplineEdge()
        if fit_points is not None:
            spline.fit_points = Vec2.list(fit_points)
        if control_points is not None:
            spline.control_points = Vec2.list(control_points)
        if knot_values is not None:
            spline.knot_values = list(knot_values)
        else:
            spline.knot_values = list(
                open_uniform_knot_vector(len(spline.control_points), degree + 1)
            )
        if weights is not None:
            spline.weights = list(weights)
        spline.degree = degree
        spline.rational = int(bool(len(spline.weights)))
        spline.periodic = int(periodic)
        if start_tangent is not None:
            spline.start_tangent = Vec2(start_tangent)
        if end_tangent is not None:
            spline.end_tangent = Vec2(end_tangent)
        self.edges.append(spline)
        return spline

    def add_spline_control_frame(
        self,
        fit_points: Iterable[tuple[float, float]],
        degree: int = 3,
        method: str = "distance",
    ) -> SplineEdge:
        bspline = global_bspline_interpolation(
            fit_points=fit_points, degree=degree, method=method
        )
        return self.add_spline(
            fit_points=fit_points,
            control_points=bspline.control_points,
            knot_values=bspline.knots(),
        )

    def clear(self) -> None:
        """Delete all edges."""
        self.edges = []

    def export_dxf(self, tagwriter: AbstractTagWriter, dxftype: str) -> None:
        tagwriter.write_tag2(92, int(self.path_type_flags))
        tagwriter.write_tag2(93, len(self.edges))
        for edge in self.edges:
            edge.export_dxf(tagwriter)
        export_source_boundary_objects(tagwriter, self.source_boundary_objects)

    def close_gaps(self, len_tol: float) -> None:
        """Insert line-edges between the existing edges if the gap between these edges
        are bigger than `len_tol`.

        .. versionadded:: 1.4

        """
        if len(self.edges) < 2:
            return
        current_edges = list(self.edges)
        first_edge = current_edges.pop(0)
        current_edges.append(first_edge)

        new_edges: list[AbstractEdge] = [first_edge]
        end_point = first_edge.real_end_point
        for edge in current_edges:
            start_point = edge.real_start_point
            if end_point.distance(start_point) > len_tol:
                line_edge = LineEdge()
                line_edge.start = end_point
                line_edge.end = start_point
                new_edges.append(line_edge)
            if edge is not first_edge:
                new_edges.append(edge)
            end_point = edge.real_end_point
        self.edges = new_edges


class LineEdge(AbstractEdge):
    EDGE_TYPE = "LineEdge"  # 2021-05-31: deprecated use type
    type = EdgeType.LINE

    def __init__(self):
        self.start = Vec2(0, 0)  # OCS!
        self.end = Vec2(0, 0)  # OCS!

    def is_valid(self) -> bool:
        return True

    @property
    def start_point(self) -> Vec2:
        return self.start

    @property
    def end_point(self) -> Vec2:
        return self.end

    @classmethod
    def load_tags(cls, tags: Tags) -> LineEdge:
        edge = cls()
        for tag in tags:
            code, value = tag
            if code == 10:
                edge.start = Vec2(value)
            elif code == 11:
                edge.end = Vec2(value)
        return edge

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tag2(72, 1)  # edge type

        x, y, *_ = self.start
        tagwriter.write_tag2(10, float(x))
        tagwriter.write_tag2(20, float(y))

        x, y, *_ = self.end
        tagwriter.write_tag2(11, float(x))
        tagwriter.write_tag2(21, float(y))

    def transform(self, ocs: OCSTransform, elevation: float) -> None:
        self.start = ocs.transform_2d_vertex(self.start, elevation)
        self.end = ocs.transform_2d_vertex(self.end, elevation)


class ArcEdge(AbstractEdge):
    type = EdgeType.ARC  # 2021-05-31: deprecated use type
    EDGE_TYPE = "ArcEdge"

    def __init__(self) -> None:
        self.center = Vec2(0.0, 0.0)
        self.radius: float = 1.0
        # Start- and end angles are always stored in counter-clockwise order!
        self.start_angle: float = 0.0
        self.end_angle: float = 360.0
        # Flag to preserve the required orientation for DXF export:
        self.ccw: bool = True

    def is_valid(self) -> bool:
        return True

    @property
    def start_point(self) -> Vec2:
        return self.construction_tool().start_point

    @property
    def real_start_point(self) -> Vec2:
        if self.ccw:
            return self.start_point
        return self.end_point

    @property
    def end_point(self) -> Vec2:
        return self.construction_tool().end_point

    @property
    def real_end_point(self) -> Vec2:
        if self.ccw:
            return self.end_point
        return self.start_point

    @classmethod
    def load_tags(cls, tags: Tags) -> ArcEdge:
        edge = cls()
        start = 0.0
        end = 0.0
        for tag in tags:
            code, value = tag
            if code == 10:
                edge.center = Vec2(value)
            elif code == 40:
                edge.radius = value
            elif code == 50:
                start = value
            elif code == 51:
                end = value
            elif code == 73:
                edge.ccw = bool(value)

        # The DXF format stores the clockwise oriented start- and end angles
        # for HATCH arc- and ellipse edges as complementary angle (360-angle).
        # This is a problem in many ways for processing clockwise oriented
        # angles correct, especially rotation transformation won't work.
        # Solution: convert clockwise angles into counter-clockwise angles
        # and swap start- and end angle at loading and exporting, the ccw flag
        # preserves the required orientation of the arc:
        if edge.ccw:
            edge.start_angle = start
            edge.end_angle = end
        else:
            edge.start_angle = 360.0 - end
            edge.end_angle = 360.0 - start
        return edge

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tag2(72, 2)  # edge type
        x, y, *_ = self.center
        if self.ccw:
            start = self.start_angle
            end = self.end_angle
        else:
            # swap and convert to complementary angles: see ArcEdge.load_tags()
            # for explanation
            start = 360.0 - self.end_angle
            end = 360.0 - self.start_angle
        tagwriter.write_tag2(10, float(x))
        tagwriter.write_tag2(20, float(y))
        tagwriter.write_tag2(40, self.radius)
        tagwriter.write_tag2(50, start)
        tagwriter.write_tag2(51, end)
        tagwriter.write_tag2(73, int(self.ccw))

    def transform(self, ocs: OCSTransform, elevation: float) -> None:
        self.center = ocs.transform_2d_vertex(self.center, elevation)
        self.radius = ocs.transform_length(Vec3(self.radius, 0, 0))
        if not math.isclose(
            arc_angle_span_deg(self.start_angle, self.end_angle), 360.0
        ):  # open arc
            # The transformation of the ccw flag is not necessary for the current
            # implementation of OCS transformations. The arc angles have always
            # a counter clockwise orientation around the extrusion vector and
            # this orientation is preserved even for mirroring, which flips the
            # extrusion vector to (0, 0, -1) for entities in the xy-plane.

            self.start_angle = ocs.transform_deg_angle(self.start_angle)
            self.end_angle = ocs.transform_deg_angle(self.end_angle)
        else:  # full circle
            # Transform only start point to preserve the connection point to
            # adjacent edges:
            self.start_angle = ocs.transform_deg_angle(self.start_angle)
            # ArcEdge is represented in counter-clockwise orientation:
            self.end_angle = self.start_angle + 360.0

    def construction_tool(self) -> ConstructionArc:
        """Returns ConstructionArc() for the OCS representation."""
        return ConstructionArc(
            center=self.center,
            radius=self.radius,
            start_angle=self.start_angle,
            end_angle=self.end_angle,
        )


class EllipseEdge(AbstractEdge):
    EDGE_TYPE = "EllipseEdge"  # 2021-05-31: deprecated use type
    type = EdgeType.ELLIPSE

    def __init__(self) -> None:
        self.center = Vec2((0.0, 0.0))
        # Endpoint of major axis relative to center point (in OCS)
        self.major_axis = Vec2((1.0, 0.0))
        self.ratio: float = 1.0
        # Start- and end angles are always stored in counter-clockwise order!
        self.start_angle: float = 0.0  # start param, not a real angle
        self.end_angle: float = 360.0  # end param, not a real angle
        # Flag to preserve the required orientation for DXF export:
        self.ccw: bool = True

    def is_valid(self) -> bool:
        return True

    @property
    def start_point(self) -> Vec2:
        return self.construction_tool().start_point.vec2

    @property
    def end_point(self) -> Vec2:
        return self.construction_tool().end_point.vec2

    @property
    def real_start_point(self) -> Vec2:
        if self.ccw:
            return self.start_point
        return self.end_point

    @property
    def real_end_point(self) -> Vec2:
        if self.ccw:
            return self.end_point
        return self.start_point

    @property
    def start_param(self) -> float:
        return angle_to_param(self.ratio, math.radians(self.start_angle))

    @start_param.setter
    def start_param(self, param: float) -> None:
        self.start_angle = math.degrees(param_to_angle(self.ratio, param))

    @property
    def end_param(self) -> float:
        return angle_to_param(self.ratio, math.radians(self.end_angle))

    @end_param.setter
    def end_param(self, param: float) -> None:
        self.end_angle = math.degrees(param_to_angle(self.ratio, param))

    @classmethod
    def load_tags(cls, tags: Tags) -> EllipseEdge:
        edge = cls()
        start = 0.0
        end = 0.0
        for tag in tags:
            code, value = tag
            if code == 10:
                edge.center = Vec2(value)
            elif code == 11:
                edge.major_axis = Vec2(value)
            elif code == 40:
                edge.ratio = value
            elif code == 50:
                start = value
            elif code == 51:
                end = value
            elif code == 73:
                edge.ccw = bool(value)

        if edge.ccw:
            edge.start_angle = start
            edge.end_angle = end
        else:
            # The DXF format stores the clockwise oriented start- and end angles
            # for HATCH arc- and ellipse edges as complementary angle (360-angle).
            # This is a problem in many ways for processing clockwise oriented
            # angles correct, especially rotation transformation won't work.
            # Solution: convert clockwise angles into counter-clockwise angles
            # and swap start- and end angle at loading and exporting, the ccw flag
            # preserves the required orientation of the ellipse:
            edge.start_angle = 360.0 - end
            edge.end_angle = 360.0 - start

        return edge

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tag2(72, 3)  # edge type
        x, y, *_ = self.center
        tagwriter.write_tag2(10, float(x))
        tagwriter.write_tag2(20, float(y))
        x, y, *_ = self.major_axis
        tagwriter.write_tag2(11, float(x))
        tagwriter.write_tag2(21, float(y))
        tagwriter.write_tag2(40, self.ratio)
        if self.ccw:
            start = self.start_angle
            end = self.end_angle
        else:
            # swap and convert to complementary angles: see EllipseEdge.load_tags()
            # for explanation
            start = 360.0 - self.end_angle
            end = 360.0 - self.start_angle

        tagwriter.write_tag2(50, start)
        tagwriter.write_tag2(51, end)
        tagwriter.write_tag2(73, int(self.ccw))

    def construction_tool(self) -> ConstructionEllipse:
        """Returns ConstructionEllipse() for the OCS representation."""
        return ConstructionEllipse(
            center=Vec3(self.center),
            major_axis=Vec3(self.major_axis),
            extrusion=Vec3(0, 0, 1),
            ratio=self.ratio,
            # 1. ConstructionEllipse() is always in ccw orientation
            # 2. Start- and end params are always stored in ccw orientation
            start_param=self.start_param,
            end_param=self.end_param,
        )

    def transform(self, ocs: OCSTransform, elevation: float) -> None:
        e = self.construction_tool()

        # Transform old OCS representation to WCS
        ocs_to_wcs = ocs.old_ocs.to_wcs
        e.center = ocs_to_wcs(e.center.replace(z=elevation))
        e.major_axis = ocs_to_wcs(e.major_axis)
        e.extrusion = ocs.old_extrusion

        # Apply matrix transformation
        e.transform(ocs.m)

        # Transform WCS representation to new OCS
        wcs_to_ocs = ocs.new_ocs.from_wcs
        self.center = wcs_to_ocs(e.center).vec2
        self.major_axis = wcs_to_ocs(e.major_axis).vec2
        self.ratio = e.ratio

        # ConstructionEllipse() is always in ccw orientation
        # Start- and end params are always stored in ccw orientation
        self.start_param = e.start_param
        self.end_param = e.end_param

        # The transformation of the ccw flag is not necessary for the current
        # implementation of OCS transformations.
        # An ellipse as boundary edge is an OCS entity!
        # The ellipse angles have always a counter clockwise orientation around
        # the extrusion vector and this orientation is preserved even for
        # mirroring, which flips the extrusion vector to (0, 0, -1) for
        # entities in the xy-plane.

        # normalize angles in range 0 to 360 degrees
        self.start_angle = self.start_angle % 360.0
        self.end_angle = self.end_angle % 360.0
        if math.isclose(self.end_angle, 0):
            self.end_angle = 360.0


class SplineEdge(AbstractEdge):
    EDGE_TYPE = "SplineEdge"  # 2021-05-31: deprecated use type
    type = EdgeType.SPLINE

    def __init__(self) -> None:
        self.degree: int = 3  # code = 94
        self.rational: int = 0  # code = 73
        self.periodic: int = 0  # code = 74
        self.knot_values: list[float] = []
        self.control_points: list[Vec2] = []
        self.fit_points: list[Vec2] = []
        self.weights: list[float] = []
        # do not set tangents by default to (0, 0)
        self.start_tangent: Optional[Vec2] = None
        self.end_tangent: Optional[Vec2] = None

    def is_valid(self) -> bool:
        if len(self.control_points):
            order = self.degree + 1
            count = len(self.control_points)
            if order > count:
                return False
            required_knot_count = count + order
            if len(self.knot_values) != required_knot_count:
                return False
        elif len(self.fit_points) < 2:
            return False
        return True

    @property
    def start_point(self) -> Vec2:
        return self.control_points[0]

    @property
    def end_point(self) -> Vec2:
        return self.control_points[-1]

    @classmethod
    def load_tags(cls, tags: Tags) -> SplineEdge:
        edge = cls()
        for tag in tags:
            code, value = tag
            if code == 94:
                edge.degree = value
            elif code == 73:
                edge.rational = value
            elif code == 74:
                edge.periodic = value
            elif code == 40:
                edge.knot_values.append(value)
            elif code == 42:
                edge.weights.append(value)
            elif code == 10:
                edge.control_points.append(Vec2(value))
            elif code == 11:
                edge.fit_points.append(Vec2(value))
            elif code == 12:
                edge.start_tangent = Vec2(value)
            elif code == 13:
                edge.end_tangent = Vec2(value)
        return edge

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        def set_required_tangents(points: list[Vec2]):
            if len(points) > 1:
                if self.start_tangent is None:
                    self.start_tangent = points[1] - points[0]
                if self.end_tangent is None:
                    self.end_tangent = points[-1] - points[-2]

        if len(self.weights):
            if len(self.weights) == len(self.control_points):
                self.rational = 1
            else:
                raise const.DXFValueError(
                    "SplineEdge: count of control points and count of weights "
                    "mismatch"
                )
        else:
            self.rational = 0

        write_tag = tagwriter.write_tag2
        write_tag(72, 4)  # edge type
        write_tag(94, int(self.degree))
        write_tag(73, int(self.rational))
        write_tag(74, int(self.periodic))
        write_tag(95, len(self.knot_values))  # number of knots
        write_tag(96, len(self.control_points))  # number of control points
        # build knot values list
        # knot values have to be present and valid, otherwise AutoCAD crashes
        if len(self.knot_values):
            for value in self.knot_values:
                write_tag(40, float(value))
        else:
            raise const.DXFValueError("SplineEdge: missing required knot values")

        # build control points
        # control points have to be present and valid, otherwise AutoCAD crashes
        cp = Vec2.list(self.control_points)
        if self.rational:
            for point, weight in zip(cp, self.weights):
                write_tag(10, float(point.x))
                write_tag(20, float(point.y))
                write_tag(42, float(weight))
        else:
            for x, y in cp:
                write_tag(10, float(x))
                write_tag(20, float(y))

        # build optional fit points
        if len(self.fit_points) > 0:
            set_required_tangents(cp)
            write_tag(97, len(self.fit_points))
            for x, y, *_ in self.fit_points:
                write_tag(11, float(x))
                write_tag(21, float(y))
        elif tagwriter.dxfversion >= const.DXF2010:
            # (97, 0) len tag required by AutoCAD 2010+
            write_tag(97, 0)

        if self.start_tangent is not None:
            x, y, *_ = self.start_tangent
            write_tag(12, float(x))
            write_tag(22, float(y))

        if self.end_tangent is not None:
            x, y, *_ = self.end_tangent
            write_tag(13, float(x))
            write_tag(23, float(y))

    def transform(self, ocs: OCSTransform, elevation: float) -> None:
        self.control_points = list(
            ocs.transform_2d_vertex(v, elevation) for v in self.control_points
        )
        self.fit_points = list(
            ocs.transform_2d_vertex(v, elevation) for v in self.fit_points
        )
        if self.start_tangent is not None:
            t = Vec3(self.start_tangent).replace(z=elevation)
            self.start_tangent = ocs.transform_direction(t).vec2
        if self.end_tangent is not None:
            t = Vec3(self.end_tangent).replace(z=elevation)
            self.end_tangent = ocs.transform_direction(t).vec2

    def construction_tool(self) -> BSpline:
        """Returns BSpline() for the OCS representation."""
        return BSpline(
            control_points=self.control_points,
            knots=self.knot_values,
            order=self.degree + 1,
            weights=self.weights,
        )


EDGE_CLASSES = [None, LineEdge, ArcEdge, EllipseEdge, SplineEdge]



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.render import point
from ezdxf.entities import Point
from ezdxf.math.shape import Shape2d


def pnt(location=(0, 0), angle: float = 0):
    return Point.new(
        dxfattribs={
            "angle": angle,
            "location": location,
        }
    )


def test_dimensionless_point():
    loc = (2, 3)
    p = pnt(location=loc)
    result = point.virtual_entities(p, pdmode=0)
    line = result[0]
    assert line.dxftype() == "LINE"
    assert line.dxf.start.isclose(loc)
    assert line.dxf.end.isclose(loc)


def test_none_point():
    p = pnt()
    result = point.virtual_entities(p, pdmode=1)
    assert len(result) == 0


def test_cross_point():
    p = pnt()
    result = point.virtual_entities(p, pdmode=2)
    line1, line2 = result
    assert line1.dxf.start == (-1, 0)
    assert line1.dxf.end == (+1, 0)
    assert line2.dxf.start == (0, -1)
    assert line2.dxf.end == (0, +1)


def test_x_cross_point():
    p = pnt()
    result = point.virtual_entities(p, pdmode=3)
    line1, line2 = result
    assert line1.dxf.start == (-1, -1)
    assert line1.dxf.end == (+1, +1)
    assert line2.dxf.start == (-1, +1)
    assert line2.dxf.end == (+1, -1)


def test_tick_point():
    p = pnt()
    result = point.virtual_entities(p, pdmode=4)
    line1 = result[0]
    assert line1.dxf.start == (0, 0)
    assert line1.dxf.end == (0, 0.5)


def test_square_point():
    p = pnt()
    result = point.virtual_entities(p, pdmode=65)
    line1, line2, line3, line4 = result
    lower_left = (-0.5, -0.5)
    assert line1.dxf.start == lower_left
    lower_right = (0.5, -0.5)
    assert line1.dxf.end == lower_right
    assert line2.dxf.start == lower_right
    upper_right = (0.5, 0.5)
    assert line2.dxf.end == upper_right
    assert line3.dxf.start == upper_right
    upper_left = (-0.5, 0.5)
    assert line3.dxf.end == upper_left
    assert line4.dxf.start == upper_left
    assert line4.dxf.end == lower_left


def test_circle_point():
    p = pnt()
    result = point.virtual_entities(p, pdmode=33)
    circle = result[0]
    assert circle.dxf.center == (0, 0)
    assert circle.dxf.radius == 0.5


def test_rotated_cross_point():
    expected = Shape2d([(-1, 0), (1, 0), (0, -1), (0, 1)])
    expected.rotate(-30)  # count-clockwise rotation
    s1, e1, s2, e2 = expected.vertices

    p = pnt(angle=30)  # clockwise angle!!
    result = point.virtual_entities(p, pdmode=2)
    line1, line2 = result
    assert line1.dxf.start.isclose(s1)
    assert line1.dxf.end.isclose(e1)
    assert line2.dxf.start.isclose(s2)
    assert line2.dxf.end.isclose(e2)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.units import DrawingUnits, PaperSpaceUnits, conversion_factor, unit_name
from math import isclose


def test_msp_1_m():
    # 1 drawing unit = 1 m
    u = DrawingUnits(1, "m")
    assert 7 * u("m") == 7, "should be 7 drawing units"
    # rounding errors!
    assert isclose(7 * u("dm"), 0.7), "should be 0.7 drawing units"
    assert 7 * u("cm") == 0.07, "should be 0.07 drawing units"
    assert 7 * u("mm") == 0.007, "should be 0.007 drawing units"
    assert 7 * u("km") == 7000, "should be 7000 drawing units"
    assert isclose(7 * u("ft"), 2.1336), "should be 2.1336 drawing units"
    assert isclose(70 * u("in"), 1.778), "should be 1.778 drawing units"
    assert isclose(7 * u("yd"), 6.4008), "should be 6.4008 drawing units"


def test_msp_easy_usage():
    u = DrawingUnits(1, "m")
    cm = u("cm")
    assert 100 * cm == 1


def test_msp_2_m():
    # 1 drawing unit = 2 m
    u = DrawingUnits(2, "m")
    assert 1 * u("m") == 2, "should be 2 drawing units"
    assert 7 * u("m") == 14, "should be 14 drawing units"
    assert 7 * u("cm") == 0.14, "should be 0.14 drawing units"
    assert 7 * u("mm") == 0.014, "should be 0.014 drawing units"
    assert 7 * u("km") == 14000, "should be 14000 drawing units"


def test_msp_1_ft():
    # 1 drawing unit = 1 ft
    u = DrawingUnits(1, "ft")
    assert isclose(1 * u("ft"), 1), "should be 1 drawing unit"
    assert isclose(10 * u("yd"), 30), "should be 30 drawing unit"
    assert isclose(
        100 * u("in"), 8.3333333333
    ), "should be 8.33333 drawing unit"
    assert round(u("mi")) == 5280, "should be 5280 drawing unit"
    assert isclose(100 * u("m"), 328.0839895), "should be 328.084 drawing unit"
    assert isclose(u("km"), 3280.839895), "should be 3280.8399 drawing unit"


def test_msp_2_ft():
    # 1 drawing unit = 2 ft
    u = DrawingUnits(2, "ft")
    assert isclose(1 * u("ft"), 2), "should be 2 drawing unit"
    assert isclose(10 * u("yd"), 60), "should be 60 drawing unit"


def test_msp_1_in():
    # 1 drawing unit = 1 in
    u = DrawingUnits(1, "in")
    assert isclose(1 * u("ft"), 12), "should be 12 drawing units"
    assert isclose(1 * u("yd"), 36), "should be 36 drawing units"
    assert round(u("mi")) == 63360, "should be 63360 drawing units"
    assert isclose(
        100 * u("m"), 3937.007874
    ), "should be 3937.0079 drawing units"
    assert isclose(u("km"), 39370.07874)
    assert isclose(100 * u("dm"), 393.7007874)
    assert isclose(100 * u("cm"), 39.37007874)
    assert isclose(100 * u("mm"), 3.937007874)
    assert isclose(100 * u("µm"), 0.003937007874)


def test_paper_space_scale_1_1():
    psu = PaperSpaceUnits(DrawingUnits(1, "m"), unit="mm", scale=1.0)
    assert psu.from_msp(1, "m") == 1000, "should be 1000 mm in paper space"
    assert psu.from_msp(10, "cm") == 100
    assert psu.from_msp(10, "mm") == 10

    psu = PaperSpaceUnits(DrawingUnits(1, "yd"), unit="in", scale=1.0)
    assert (
        round(psu.from_msp(1, "yd")) == 36
    ), "should be 36 inch in paper space"
    assert (
        round(psu.from_msp(1, "ft")) == 12
    ), "should be 12 inch in paper space"


def test_paper_space_scale_1_100():
    psu = PaperSpaceUnits(DrawingUnits(1, "m"), unit="mm", scale=100.0)
    assert (
        psu.from_msp(1, "m") == 10
    ), "1m in msp, should be 10 mm in paper space 1:100"
    assert psu.from_msp(500, "cm") == 50

    psu = PaperSpaceUnits(DrawingUnits(1, "yd"), unit="in", scale=100.0)
    assert (
        round(psu.from_msp(1, "yd"), 2) == 0.36
    ), "1 yd in msp, should be .36 inch in paper space 1:100"
    assert (
        round(psu.from_msp(1, "ft"), 2) == 0.12
    ), "1 ft in msp, should be .12 inch in paper space 1:100"


def test_paper_space_default():
    # Default: 1 drawing unit = 1 m, paperspace units = 'mm', scale = 1:1
    psu = PaperSpaceUnits()
    assert psu.from_msp(1, "m") == 1000

    psu = PaperSpaceUnits(scale=100)
    assert psu.from_msp(1, "m") == 10


def test_paper_space_to_model_space_metric():
    psu = PaperSpaceUnits(msp=DrawingUnits(1, "m"), scale=100)
    assert (
        psu.to_msp(10, "mm") == 1
    ), "1mm in PSP 1:100 => should be 1 drawing unit in model space"
    assert (
        psu.to_msp(5, "cm") == 5
    ), "5cm in PSP 1:100 => should be 5 drawing unit in model space"


def test_paper_space_to_model_space_us():
    psu = PaperSpaceUnits(msp=DrawingUnits(1, "yd"), scale=100)
    assert isclose(
        psu.to_msp(1, "in"), 2.77777777777
    ), "1 inch in PSP 1:100 => should be 2.777 drawing unit in model space"


@pytest.mark.parametrize(
    "source, target, factor",
    [
        [6, 4, 1000],  # m in mm
        [4, 6, 0.001],  # mm in m
        [6, 5, 100],  # m in cm
        [5, 6, 0.01],  # cm in m
        [5, 4, 10],  # cm in mm
        [4, 5, 0.1],  # mm in cm
    ],
)
def test_conversion_factor(source, target, factor):
    assert conversion_factor(source, target) == pytest.approx(factor)


@pytest.mark.parametrize(
    "source, target",
    [
        [0, 6],  # 0 = unitless
        [6, 0],
        [0, 0],
    ],
)
def test_conversion_type_error(source, target):
    with pytest.raises(TypeError):
        conversion_factor(source, target)


def test_conversion_invalid_units():
    with pytest.raises(ValueError):
        conversion_factor(25, 6)


def test_unit_name_valid_input():
    assert unit_name(0) == "Unitless"
    assert unit_name(1) == "Inches"


def test_unit_name_invalid_input():
    assert unit_name(-1) == "unitless"



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture
def src():
    doc = ezdxf.new()
    doc.layers.new("EXTRA_LAYER")
    doc.linetypes.new("EXTRA_LT")
    doc.styles.new("EXTRA_STYLE")
    attribs = {
        "linetype": "EXTRA_LT",
        "layer": "EXTRA_LAYER",
    }
    msp = doc.modelspace()

    msp.add_line((0, 0), (1, 0), dxfattribs=attribs)
    msp.add_text(
        "Text",
        dxfattribs={
            "layer": "EXTRA_LAYER",
            "style": "EXTRA_STYLE",
        },
    )
    msp.add_polyline2d([(0, 0), (1, 1), (2, 2)], dxfattribs=attribs)
    return msp


@pytest.fixture
def target():
    doc = ezdxf.new()
    # for different handles
    for x in range(10):
        doc.entitydb.next_handle()
    return doc.modelspace()


def test_01_copy_foreign_entities_reset_resources(src, target):
    src_length = len(src)
    line = src[0]
    assert line.dxftype() == "LINE"
    target.add_foreign_entity(line, copy=True)
    line_copy = target[0]
    assert line_copy.dxftype() == "LINE"
    # check if copied not moved
    assert len(src) == src_length
    assert len(target) == 1
    assert line_copy.doc is not None
    assert line_copy.doc is not line.doc
    assert line_copy.dxf.handle != line.dxf.handle
    # check resources
    # layer is preserved
    assert line_copy.dxf.layer == line.dxf.layer
    assert line_copy.dxf.linetype == "BYLAYER"


def test_02_copy_foreign_entities_with_resources(src, target):
    line = src[0]
    assert line.dxftype() == "LINE"
    assert line.dxf.layer != "0"
    # create source layer
    target.doc.layers.new(line.dxf.layer)
    target.doc.linetypes.new(line.dxf.linetype)
    target.add_foreign_entity(line, copy=True)
    line_copy = target[0]
    assert line_copy.dxftype() == "LINE"
    # check resources
    assert line_copy.dxf.layer == line.dxf.layer
    assert line_copy.dxf.linetype == line.dxf.linetype


def test_03_move_foreign_entities(src, target):
    src_length = len(src)
    text = src[1]
    assert text.dxftype() == "TEXT"
    target.add_foreign_entity(text, copy=False)
    text_moved = target[0]
    assert text_moved.dxftype() == "TEXT"

    # check if moved not copied
    assert len(src) == src_length - 1
    assert len(target) == 1
    assert text_moved is text
    assert text_moved.doc is not None

    # check resources
    assert text_moved.dxf.layer == "EXTRA_LAYER", "layer should be preserved"
    assert text_moved.dxf.linetype == "BYLAYER"
    assert text_moved.dxf.style == "Standard"


def test_04_move_foreign_entities_with_resources(src, target):
    src_length = len(src)
    text = src[1]
    assert text.dxftype() == "TEXT"
    layer = text.dxf.layer
    style = text.dxf.style
    target.doc.layers.new(layer)
    target.doc.styles.new(style)

    target.add_foreign_entity(text, copy=False)
    text_moved = target[0]
    assert text_moved.dxftype() == "TEXT"

    # check if moved not copied
    assert len(src) == src_length - 1
    assert len(target) == 1
    assert text_moved is text
    assert text_moved.doc is not None

    # check resources
    assert text_moved.dxf.layer == layer
    assert text_moved.dxf.linetype == "BYLAYER"
    assert text_moved.dxf.style == style


def test_05_copy_polyline_reset_resources(src, target):
    src_length = len(src)
    pline = src[2]
    assert pline.dxftype() == "POLYLINE"
    target.add_foreign_entity(pline, copy=True)
    pline_copy = target[0]
    assert pline_copy.dxftype() == "POLYLINE"
    # check if copied not moved
    assert len(src) == src_length
    assert len(target) == 1
    assert pline_copy.doc is not None
    assert pline_copy.doc is not pline.doc
    assert pline_copy.dxf.handle != pline.dxf.handle
    # check resources
    assert pline_copy.dxf.layer == "EXTRA_LAYER", "layer should be preserved"
    assert pline_copy.dxf.linetype == "BYLAYER"
    # check vertices
    assert len(pline_copy.vertices) == len(pline.vertices)
    for v1, v2 in zip(pline_copy.vertices, pline.vertices):
        assert v1.dxf.handle != v2.dxf.handle
        assert v1.dxf.layer == "EXTRA_LAYER", "layer should be preserved"
        assert v1.dxf.linetype != v2.dxf.linetype


def test_06_move_polyline(src, target):
    src_length = len(src)
    pline = src[2]
    assert pline.dxftype() == "POLYLINE"
    handles = [v.dxf.handle for v in pline.vertices]
    handles.append(pline.dxf.handle)

    target.add_foreign_entity(pline, copy=False)
    pline_move = target[0]
    assert pline_move.dxftype() == "POLYLINE"
    # check if moved
    assert len(src) == src_length - 1
    assert len(target) == 1
    assert pline_move.doc is not None
    assert pline_move.doc is pline.doc

    # check removed entities
    for handle in handles:
        assert handle not in src.doc.entitydb


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
"""
Test the ezdxf.entities.objectcollection module, but the tests need
a real implementation: MLeaderStyleCollection

"""
import pytest

import ezdxf
from ezdxf.entities import is_dxf_object


@pytest.fixture(scope="module")
def collection_ro():
    """Creates a read only document"""
    doc = ezdxf.new()
    doc.entitydb.locked = True
    return doc.mleader_styles


class TestGetterMethods:
    def test_len(self, collection_ro):
        assert len(collection_ro) == 1

    def test_iter(self, collection_ro):
        assert len(list(collection_ro)) == 1

    def test_is_unique_name(self, collection_ro):
        assert collection_ro.is_unique_name("STANDARD") is False

    def test_contains(self, collection_ro):
        assert ("Standard" in collection_ro) is True

    def test_contains_is_case_insensitive(self, collection_ro):
        assert ("STANDARD" in collection_ro) is True

    def test_getitem(self, collection_ro):
        assert collection_ro["Standard"].dxf.name == "Standard"

    def test_getitem_is_case_insensitive(self, collection_ro):
        assert collection_ro["STANDARD"].dxf.name == "Standard"

    def test_get(self, collection_ro):
        assert collection_ro.get("Standard").dxf.name == "Standard"

    def test_get_is_case_insensitive(self, collection_ro):
        assert collection_ro.get("STANDARD").dxf.name == "Standard"


@pytest.fixture(scope="module")
def collection_rw():
    doc = ezdxf.new()
    return doc.mleader_styles


class TestCreateNewEntry:
    def test_new_entry_is_an_object(self, collection_rw):
        obj = collection_rw.new("New1")
        assert is_dxf_object(obj) is True
        assert obj.dxf.name == "New1"

    def test_new_entry_is_added_to_collection(self, collection_rw):
        count = len(collection_rw)
        collection_rw.new("New2")
        assert len(collection_rw) == count + 1
        assert "NEW2" in collection_rw, "case insensitive names"

    def test_cannot_use_existing_name(self, collection_rw):
        collection_rw.new("New3")
        with pytest.raises(ValueError):
            collection_rw.new("NEW3"), "case insensitive names"

    def test_invalid_char_in_name_raises_exception(self, collection_rw):
        with pytest.raises(ValueError):
            collection_rw.new("New:")


class TestDeleteEntry:
    def test_delete_entry_remove_entry(self, collection_rw):
        count = len(collection_rw)
        collection_rw.new("Del1")
        collection_rw.delete("DEL1")
        assert len(collection_rw) == count

    def test_delete_non_existing_entry_does_not_raise_exception(
        self, collection_rw
    ):
        count = len(collection_rw)
        collection_rw.delete("DEL2")
        assert len(collection_rw) == count


class TestDuplicateEntry:
    def test_duplicate_existing_entry(self, collection_rw):
        count = len(collection_rw)
        obj = collection_rw.duplicate_entry("STANDARD", "Dup1")
        assert is_dxf_object(obj) is True
        assert obj.dxf.name == "Dup1"
        assert len(collection_rw) == count + 1

    def test_duplicate_non_existing_entry_raises_exception(self, collection_rw):
        with pytest.raises(ValueError):
            collection_rw.duplicate_entry("NON_EXISTING_ENTRY", "Dup1")

    def test_new_entry_replaces_existing_entry(self, collection_rw):
        count = len(collection_rw)
        obj1 = collection_rw.duplicate_entry("STANDARD", "Dup2")
        obj2 = collection_rw.duplicate_entry("Standard", "DUP2")
        assert obj1 is not obj2, "obj2 must be a new object"
        assert collection_rw.get("Dup2") is obj2, "obj2 should replace obj1"
        assert len(collection_rw) == count + 1

    def test_duplicated_entries_have_same_content(self, collection_rw):
        def attribs(obj):
            a = obj.dxf.all_existing_dxf_attribs()
            del a['handle']
            del a['name']
            return a

        obj0 = collection_rw.get("Standard")
        obj1 = collection_rw.duplicate_entry("STANDARD", "Dup3")
        attribs0 = attribs(obj0)
        attribs1 = attribs(obj1)
        assert len(attribs0) > 1
        assert attribs0 == attribs1
        assert obj1.get_reactors() == obj0.get_reactors()

    def test_duplicated_entry_is_stored_in_objects_section(self, collection_rw):
        obj = collection_rw.duplicate_entry("STANDARD", "Dup4")
        assert obj.dxf.handle in obj.doc.objects

    def test_invalid_char_in_new_name_raises_exception(self, collection_rw):
        with pytest.raises(ValueError):
            collection_rw.duplicate_entry("Standard", "New:")


def test_clear():
    doc = ezdxf.new()
    doc.mleader_styles.clear()
    # This creates an invalid DXF file!!!
    assert len(doc.mleader_styles) == 0


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
from ezdxf.lldxf.const import SUBCLASS_MARKER
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    group_code_mapping,
)
from ezdxf.lldxf.tags import Tags
from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .factory import register_entity
from .objectcollection import ObjectCollection
from ezdxf.math import Matrix44
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace, DXFEntity
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.document import Drawing

__all__ = ["Material", "MaterialCollection"]


def fetch_matrix(tags: Tags, code: int) -> tuple[Tags, Optional[Matrix44]]:
    values = []
    remaining = Tags()
    for tag in tags:
        if tag.code == code:
            values.append(tag.value)
            if len(values) == 16:
                # enough values collected, code 43 is maybe used for two matrices
                code = -1
        else:
            remaining.append(tag)
    if len(values) == 16:
        # only if valid matrix
        return remaining, Matrix44(values)
    else:
        return tags, None


def export_matrix(tagwriter: AbstractTagWriter, code: int, matrix: Matrix44) -> None:
    if matrix is not None:
        for value in matrix:
            tagwriter.write_tag2(code, value)


acdb_material = DefSubclass(
    "AcDbMaterial",
    {
        "name": DXFAttr(1),
        "description": DXFAttr(2, default=""),
        "ambient_color_method": DXFAttr(
            70, default=0
        ),  # 0=use current color; 1=override current color
        "ambient_color_factor": DXFAttr(40, default=1.0),  # valid range is 0.0 to 1.0
        "ambient_color_value": DXFAttr(90),  # integer representing an AcCmEntityColor
        "diffuse_color_method": DXFAttr(
            71, default=0
        ),  # 0=use current color; 1=override current color
        "diffuse_color_factor": DXFAttr(41, default=1.0),  # valid range is 0.0 to 1.0
        "diffuse_color_value": DXFAttr(
            91, default=-1023410177
        ),  # integer representing an AcCmEntityColor
        "diffuse_map_blend_factor": DXFAttr(
            42, default=1.0
        ),  # valid range is 0.0 to 1.0
        "diffuse_map_source": DXFAttr(72, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "diffuse_map_file_name": DXFAttr(3, default=""),
        "diffuse_map_projection_method": DXFAttr(
            73, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "diffuse_map_tiling_method": DXFAttr(74, default=1),  # 1=Tile; 2=Crop; 3=Clamp
        "diffuse_map_auto_transform_method": DXFAttr(75, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 43: Transform matrix of diffuse map mapper (16 reals; row major format; default = identity matrix)
        "specular_gloss_factor": DXFAttr(44, default=0.5),  # valid range is 0.0 to 1.0
        "specular_color_method": DXFAttr(
            73, default=0
        ),  # 0=use current color; 1=override current color
        "specular_color_factor": DXFAttr(45, default=1.0),  # valid range is 0.0 to 1.0
        "specular_color_value": DXFAttr(92),  # integer representing an AcCmEntityColor
        "specular_map_blend_factor": DXFAttr(
            46, default=1.0
        ),  # valid range is 0.0 to 1.0
        "specular_map_source": DXFAttr(77, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "specular_map_file_name": DXFAttr(4, default=""),
        "specular_map_projection_method": DXFAttr(
            78, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "specular_map_tiling_method": DXFAttr(79, default=1),  # 1=Tile; 2=Crop; 3=Clamp
        "specular_map_auto_transform_method": DXFAttr(170, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 47: Transform matrix of specular map mapper (16 reals; row major format; default = identity matrix)
        "reflection_map_blend_factor": DXFAttr(
            48, default=1.0
        ),  # valid range is 0.0 to 1.0
        "reflection_map_source": DXFAttr(171, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "reflection_map_file_name": DXFAttr(6, default=""),
        "reflection_map_projection_method": DXFAttr(
            172, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "reflection_map_tiling_method": DXFAttr(
            173, default=1
        ),  # 1=Tile; 2=Crop; 3=Clamp
        "reflection_map_auto_transform_method": DXFAttr(174, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 49: Transform matrix of reflection map mapper (16 reals; row major format; default = identity matrix)
        "opacity": DXFAttr(140, default=1.0),  # valid range is 0.0 to 1.0
        "opacity_map_blend_factor": DXFAttr(
            141, default=1.0
        ),  # valid range is 0.0 to 1.0
        "opacity_map_source": DXFAttr(175, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "opacity_map_file_name": DXFAttr(7, default=""),
        "opacity_map_projection_method": DXFAttr(
            176, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "opacity_map_tiling_method": DXFAttr(177, default=1),  # 1=Tile; 2=Crop; 3=Clamp
        "opacity_map_auto_transform_method": DXFAttr(178, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 142: Transform matrix of reflection map mapper (16 reals; row major format; default = identity matrix)
        "bump_map_blend_factor": DXFAttr(143, default=1.0),  # valid range is 0.0 to 1.0
        "bump_map_source": DXFAttr(179, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "bump_map_file_name": DXFAttr(8, default=""),
        "bump_map_projection_method": DXFAttr(
            270, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "bump_map_tiling_method": DXFAttr(271, default=1),  # 1=Tile; 2=Crop; 3=Clamp
        "bump_map_auto_transform_method": DXFAttr(272, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 144: Transform matrix of bump map mapper (16 reals; row major format; default = identity matrix)
        "refraction_index": DXFAttr(145, default=1.0),  # valid range is 0.0 to 1.0
        "refraction_map_blend_factor": DXFAttr(
            146, default=1.0
        ),  # valid range is 0.0 to 1.0
        "refraction_map_source": DXFAttr(273, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "refraction_map_file_name": DXFAttr(9, default=""),
        "refraction_map_projection_method": DXFAttr(
            274, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "refraction_map_tiling_method": DXFAttr(
            275, default=1
        ),  # 1=Tile; 2=Crop; 3=Clamp
        "refraction_map_auto_transform_method": DXFAttr(276, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 147: Transform matrix of reflection map mapper (16 reals; row major format; default = identity matrix)
        # normal map shares group codes with diffuse map
        "normal_map_method": DXFAttr(271),
        "normal_map_strength": DXFAttr(465),
        "normal_map_blend_factor": DXFAttr(
            42, default=1.0
        ),  # valid range is 0.0 to 1.0
        "normal_map_source": DXFAttr(72, default=1),
        # 0=use current scene; 1=use image file (specified by file name; null file name specifies no map)
        "normal_map_file_name": DXFAttr(3, default=""),
        "normal_map_projection_method": DXFAttr(
            73, default=1
        ),  # 1=Planar; 2=Box; 3=Cylinder; 4=Sphere
        "normal_map_tiling_method": DXFAttr(74, default=1),  # 1=Tile; 2=Crop; 3=Clamp
        "normal_map_auto_transform_method": DXFAttr(75, default=1),  # bitset;
        # 1 = No auto transform
        # 2 = Scale mapper to current entity extents; translate mapper to entity origin
        # 4 = Include current block transform in mapper transform
        # 16x group code 43: Transform matrix of reflection map mapper (16 reals; row major format; default = identity matrix)
        "color_bleed_scale": DXFAttr(460),
        "indirect_dump_scale": DXFAttr(461),
        "reflectance_scale": DXFAttr(462),
        "transmittance_scale": DXFAttr(463),
        "two_sided_material": DXFAttr(290),
        "luminance": DXFAttr(464),
        "luminance_mode": DXFAttr(270),  # multiple usage of group code 270
        "materials_anonymous": DXFAttr(293),
        "global_illumination_mode": DXFAttr(272),
        "final_gather_mode": DXFAttr(273),
        "gen_proc_name": DXFAttr(300),
        "gen_proc_val_bool": DXFAttr(291),
        "gen_proc_val_int": DXFAttr(271),
        "gen_proc_val_real": DXFAttr(469),
        "gen_proc_val_text": DXFAttr(301),
        "gen_proc_table_end": DXFAttr(292),
        "gen_proc_val_color_index": DXFAttr(62),
        "gen_proc_val_color_rgb": DXFAttr(420),
        "gen_proc_val_color_name": DXFAttr(430),
        "map_utile": DXFAttr(270),  # multiple usage of group code 270
        "translucence": DXFAttr(148),
        "self_illumination": DXFAttr(90),
        "reflectivity": DXFAttr(468),
        "illumination_model": DXFAttr(93),
        "channel_flags": DXFAttr(94, default=63),
    },
)
acdb_material_group_codes = group_code_mapping(acdb_material)


@register_entity
class Material(DXFObject):
    DXFTYPE = "MATERIAL"
    DEFAULT_ATTRIBS = {
        "diffuse_color_method": 1,
        "diffuse_color_value": -1023410177,
    }
    DXFATTRIBS = DXFAttributes(base_class, acdb_material)

    def __init__(self) -> None:
        super().__init__()
        self.diffuse_mapper_matrix: Optional[Matrix44] = None  # code 43
        self.specular_mapper_matrix: Optional[Matrix44] = None  # code 47
        self.reflexion_mapper_matrix: Optional[Matrix44] = None  # code 49
        self.opacity_mapper_matrix: Optional[Matrix44] = None  # group 142
        self.bump_mapper_matrix: Optional[Matrix44] = None  # group 144
        self.refraction_mapper_matrix: Optional[Matrix44] = None  # code 147
        self.normal_mapper_matrix: Optional[Matrix44] = None  # code 43 ???

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy material mapper matrices"""

        def copy(matrix):
            return None if matrix is None else matrix.copy()

        assert isinstance(entity, Material)
        entity.diffuse_mapper_matrix = copy(self.diffuse_mapper_matrix)
        entity.specular_mapper_matrix = copy(self.specular_mapper_matrix)
        entity.reflexion_mapper_matrix = copy(self.reflexion_mapper_matrix)
        entity.opacity_mapper_matrix = copy(self.opacity_mapper_matrix)
        entity.bump_mapper_matrix = copy(self.bump_mapper_matrix)
        entity.refraction_mapper_matrix = copy(self.refraction_mapper_matrix)
        entity.normal_mapper_matrix = copy(self.normal_mapper_matrix)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.fast_load_dxfattribs(
                dxf, acdb_material_group_codes, 1, log=False
            )
            self.load_matrices(tags)
        return dxf

    def load_matrices(self, tags):
        tags, matrix = fetch_matrix(tags, 43)
        if matrix:
            self.diffuse_mapper_matrix = matrix
        tags, matrix = fetch_matrix(tags, 47)
        if matrix:
            self.specular_mapper_matrix = matrix
        tags, matrix = fetch_matrix(tags, 49)
        if matrix:
            self.reflexion_mapper_matrix = matrix
        tags, matrix = fetch_matrix(tags, 142)
        if matrix:
            self.opacity_mapper_matrix = matrix
        tags, matrix = fetch_matrix(tags, 144)
        if matrix:
            self.bump_mapper_matrix = matrix
        tags, matrix = fetch_matrix(tags, 147)
        if matrix:
            self.refraction_mapper_matrix = matrix
        tags, matrix = fetch_matrix(tags, 43)
        if matrix:
            self.normal_mapper_matrix = matrix

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)

        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_material.name)

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "description",
                "ambient_color_method",
                "ambient_color_factor",
                "ambient_color_value",
                "diffuse_color_method",
                "diffuse_color_factor",
                "diffuse_color_value",
                "diffuse_map_blend_factor",
                "diffuse_map_source",
                "diffuse_map_file_name",
                "diffuse_map_projection_method",
                "diffuse_map_tiling_method",
                "diffuse_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 43, self.diffuse_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "specular_gloss_factor",
                "specular_color_method",
                "specular_color_factor",
                "specular_color_value",
                "specular_map_blend_factor",
                "specular_map_source",
                "specular_map_file_name",
                "specular_map_projection_method",
                "specular_map_tiling_method",
                "specular_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 47, self.specular_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "reflection_map_blend_factor",
                "reflection_map_source",
                "reflection_map_file_name",
                "reflection_map_projection_method",
                "reflection_map_tiling_method",
                "reflection_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 49, self.reflexion_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "opacity",
                "opacity_map_blend_factor",
                "opacity_map_source",
                "opacity_map_file_name",
                "opacity_map_projection_method",
                "opacity_map_tiling_method",
                "opacity_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 142, self.opacity_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "bump_map_blend_factor",
                "bump_map_source",
                "bump_map_file_name",
                "bump_map_projection_method",
                "bump_map_tiling_method",
                "bump_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 144, self.bump_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "refraction_index",
                "refraction_map_blend_factor",
                "refraction_map_source",
                "refraction_map_file_name",
                "refraction_map_projection_method",
                "refraction_map_tiling_method",
                "refraction_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 147, self.refraction_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "normal_map_method",
                "normal_map_strength",
                "normal_map_blend_factor",
                "normal_map_source",
                "normal_map_file_name",
                "normal_map_projection_method",
                "normal_map_tiling_method",
                "normal_map_auto_transform_method",
            ],
        )
        export_matrix(tagwriter, 43, self.normal_mapper_matrix)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "color_bleed_scale",
                "indirect_dump_scale",
                "reflectance_scale",
                "transmittance_scale",
                "two_sided_material",
                "luminance",
                "luminance_mode",
                "materials_anonymous",
                "global_illumination_mode",
                "final_gather_mode",
                "gen_proc_name",
                "gen_proc_val_bool",
                "gen_proc_val_int",
                "gen_proc_val_real",
                "gen_proc_val_text",
                "gen_proc_table_end",
                "gen_proc_val_color_index",
                "gen_proc_val_color_rgb",
                "gen_proc_val_color_name",
                "map_utile",
                "translucence",
                "self_illumination",
                "reflectivity",
                "illumination_model",
                "channel_flags",
            ],
        )


class MaterialCollection(ObjectCollection[Material]):
    def __init__(self, doc: Drawing):
        super().__init__(doc, dict_name="ACAD_MATERIAL", object_type="MATERIAL")
        self.create_required_entries()

    def create_required_entries(self) -> None:
        for name in ("ByBlock", "ByLayer", "Global"):
            if name not in self:
                self.new(name)



# Copyright (C) 2011-2023, Manfred Moitzi
# License: MIT License
"""Ezdxf is an interface library for the DXF file format.

The package is designed to facilitate the creation and manipulation of DXF
documents, with compatibility across various DXF versions. It empowers users to
seamlessly load and edit DXF files while preserving all content, except for comments.

Any unfamiliar DXF tags encountered in the document are gracefully ignored but retained
for future modifications. This feature enables the processing of DXF documents
containing data from third-party applications without any loss of valuable information.
"""
from typing import TextIO, Optional
import sys
import os
from .version import version, __version__

VERSION = __version__
__author__ = "mozman <me@mozman.at>"

TRUE_STATE = {"True", "true", "On", "on", "1"}
PYPY = hasattr(sys, "pypy_version_info")
PYPY_ON_WINDOWS = sys.platform.startswith("win") and PYPY

# name space imports - do not remove
from ezdxf._options import options, config_files
from ezdxf.colors import (
    int2rgb,
    rgb2int,
    transparency2float,
    float2transparency,
)
from ezdxf.enums import InsertUnits
from ezdxf.lldxf import const
from ezdxf.lldxf.validator import is_dxf_file, is_dxf_stream
from ezdxf.filemanagement import readzip, new, read, readfile, decode_base64
from ezdxf.tools.standards import (
    setup_linetypes,
    setup_styles,
    setup_dimstyles,
    setup_dimstyle,
)
from ezdxf.tools import pattern
from ezdxf.render.arrows import ARROWS
from ezdxf.lldxf.const import (
    DXFError,
    DXFStructureError,
    DXFVersionError,
    DXFTableEntryError,
    DXFAppDataError,
    DXFXDataError,
    DXFAttributeError,
    DXFValueError,
    DXFKeyError,
    DXFIndexError,
    DXFTypeError,
    DXFBlockInUseError,
    InvalidGeoDataException,
    DXF12,
    DXF2000,
    DXF2004,
    DXF2007,
    DXF2010,
    DXF2013,
    DXF2018,
)

# name space imports - do not remove

import codecs
from ezdxf.lldxf.encoding import (
    dxf_backslash_replace,
    has_dxf_unicode,
    decode_dxf_unicode,
)


# setup DXF unicode encoder -> '\U+nnnn'
codecs.register_error("dxfreplace", dxf_backslash_replace)

EZDXF_TEST_FILES = options.test_files
YES_NO = {True: "yes", False: "no"}


def print_config(verbose: bool = False, stream: Optional[TextIO] = None) -> None:
    from pathlib import Path

    if stream is None:
        stream = sys.stdout
    stream.write(
        "\n".join([
            f"ezdxf {__version__} from {Path(__file__).parent}",
            f"Python version: {sys.version}",
            f"using C-extensions: {YES_NO[options.use_c_ext]}\n",
        ])
    )
    if verbose:
        stream.write("\nConfiguration:\n")
        options.write(stream)
        stream.write("\nEnvironment Variables:\n")
        for v in options.CONFIG_VARS:
            stream.write(f"{v}={os.environ.get(v, '')}\n")

        stream.write("\nLoaded Config Files:\n")
        for path in options.loaded_config_files:
            stream.write(str(path.absolute()) + "\n")



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License

import pytest
import math
import numpy as np

pytest.importorskip("ezdxf.acc.bspline")

from ezdxf.math._bspline import Basis as PyBasis, Evaluator as PyEvaluator
from ezdxf.acc.bspline import Basis as CyBasis, Evaluator as CyEvaluator
from ezdxf.math import Vec3, close_vectors

COUNT = 10
ORDER = 4
KNOTS = tuple(range(COUNT + ORDER))
WEIGHTS = [1.0, 0.7, 0.6, 0.5, 0.5, 0.5, 0.5, 0.6, 0.7, 1.0]
POINTS = [
    Vec3(0.181, 0.753, 0.15),
    Vec3(0.527, 0.944, 0.176),
    Vec3(1.326, 1.015, 0.184),
    Vec3(1.086, 0.607, 0.185),
    Vec3(1.286, 1.468, 0.255),
    Vec3(1.451, 0.596, 0.175),
    Vec3(1.282, 0.703, 0.17),
    Vec3(0.79, 0.77, 0.169),
    Vec3(1.622, 0.831, 0.172),
    Vec3(1.099, 0.922, 0.163),
]


@pytest.fixture
def t_vector():
    return list(np.linspace(0, max(KNOTS), 20))


@pytest.fixture
def py_basis():
    return PyBasis(KNOTS, ORDER, COUNT)


@pytest.fixture
def cy_basis():
    return CyBasis(KNOTS, ORDER, COUNT)


@pytest.fixture
def py_wbasis():
    return PyBasis(KNOTS, ORDER, COUNT, WEIGHTS)


@pytest.fixture
def cy_wbasis():
    return CyBasis(KNOTS, ORDER, COUNT, WEIGHTS)


def test_find_span(py_basis, cy_basis, t_vector):
    for u in t_vector:
        assert py_basis.find_span(u) == cy_basis.find_span(u)


def test_cython_knots(cy_basis):
    for _ in range(10):
        assert cy_basis.knots == KNOTS
        assert cy_basis.max_t == KNOTS[-1]


def test_basis_funcs(py_basis, cy_basis, t_vector):
    for u in t_vector:
        span1 = py_basis.find_span(u)
        p = py_basis.basis_funcs(span1, u)
        span2 = cy_basis.find_span(u)
        c = cy_basis.basis_funcs(span2, u)
        assert (
            all(math.isclose(a, b, abs_tol=1e-12) for a, b in zip(p, c)) is True
        )


def test_basis_vector(py_basis, cy_basis, t_vector):
    for u in t_vector:
        p = py_basis.basis_vector(u)
        c = list(cy_basis.basis_vector(u))
        assert (
            all(math.isclose(a, b, abs_tol=1e-12) for a, b in zip(p, c)) is True
        )


def test_weighted_basis_vector(py_wbasis, cy_wbasis, t_vector):
    for u in t_vector:
        p = py_wbasis.basis_vector(u)
        c = list(cy_wbasis.basis_vector(u))
        assert (
            all(math.isclose(a, b, abs_tol=1e-12) for a, b in zip(p, c)) is True
        )


def test_basis_funcs_derivatives(py_basis, cy_basis, t_vector):
    for u in t_vector:
        span = py_basis.find_span(u)
        pl = py_basis.basis_funcs_derivatives(span, u, 2)
        span = cy_basis.find_span(u)
        cl = cy_basis.basis_funcs_derivatives(span, u, 2)
        for p, c in zip(pl, cl):
            assert (
                all(math.isclose(a, b, abs_tol=1e-12) for a, b in
                    zip(p, c)) is True
            )


def test_weighted_basis_funcs_derivatives(py_wbasis, cy_wbasis, t_vector):
    for u in t_vector:
        span = py_wbasis.find_span(u)
        pl = py_wbasis.basis_funcs_derivatives(span, u, 2)
        span = cy_wbasis.find_span(u)
        cl = cy_wbasis.basis_funcs_derivatives(span, u, 2)
        for p, c in zip(pl, cl):
            assert (
                all(math.isclose(a, b, abs_tol=1e-12) for a, b in
                    zip(p, c)) is True
            )


@pytest.fixture
def py_eval(py_basis):
    return PyEvaluator(py_basis, POINTS)


@pytest.fixture
def cy_eval(cy_basis):
    return CyEvaluator(cy_basis, POINTS)


def test_point_evaluator(py_eval, cy_eval, t_vector):
    py_points = list(py_eval.points(t_vector))
    cy_points = list(cy_eval.points(t_vector))
    assert close_vectors(py_points, cy_points) is True


def test_derivative_evaluator(py_eval, cy_eval, t_vector):
    py_ders = list(py_eval.derivatives(t_vector, 2))
    cy_ders = list(cy_eval.derivatives(t_vector, 2))
    for d1, d2 in zip(py_ders, cy_ders):
        assert close_vectors(d1, d2) is True

@pytest.fixture
def py_weval(py_wbasis):
    return PyEvaluator(py_wbasis, POINTS)


@pytest.fixture
def cy_weval(cy_wbasis):
    return CyEvaluator(cy_wbasis, POINTS)


def test_weighted_point_evaluator(py_weval, cy_weval, t_vector):
    py_points = list(py_weval.points(t_vector))
    cy_points = list(cy_weval.points(t_vector))
    assert close_vectors(py_points, cy_points) is True


def test_weighted_derivative_evaluator(py_weval, cy_weval, t_vector):
    py_ders = list(py_weval.derivatives(t_vector, 2))
    cy_ders = list(cy_weval.derivatives(t_vector, 2))
    for d1, d2 in zip(py_ders, cy_ders):
        assert close_vectors(d1, d2) is True


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2014-2020, Manfred Moitzi
# License: MIT-License
import pytest

import ezdxf
import os

FILE = os.path.join(ezdxf.EZDXF_TEST_FILES, "ProE_AC1018.dxf")


@pytest.mark.skipif(
    not os.path.exists(FILE),
    reason=f"Skip reading ProE AC1018: test file '{FILE}' not available.",
)
def test_open_proe_ac1018():
    doc = ezdxf.readfile(FILE)
    modelspace = doc.modelspace()

    # are there entities in model space
    assert 17 == len(modelspace)

    # can you get entities
    lines = modelspace.query("LINE")
    assert 12 == len(lines)

    # is owner tag correct
    first_line = lines[0]
    assert modelspace.layout_key == first_line.dxf.owner

    # is paper space == 0
    assert 0 == first_line.dxf.paperspace



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest
import math

import ezdxf
from ezdxf.entities.text import Text
from ezdxf.enums import TextEntityAlignment
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Vec3, Matrix44

TEST_CLASS = Text
TEST_TYPE = "TEXT"

ENTITY_R12 = """0
TEXT
5
0
8
0
10
0.0
20
0.0
30
0.0
40
1.0
1
TEXTCONTENT
50
0.0
51
0.0
7
Standard
41
1.0
71
0
72
0
11
0.0
21
0.0
31
0.0
73
0
"""

ENTITY_R2000 = """0
TEXT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbText
10
0.0
20
0.0
30
0.0
40
1.0
1
TEXTCONTENT
50
0.0
51
0.0
7
Standard
41
1.0
71
0
72
0
11
0.0
21
0.0
31
0.0
100
AcDbText
73
0
"""


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "insert": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"
    assert entity.dxf.insert == (1, 2, 3)
    assert entity.dxf.insert.x == 1, "is not Vec3 compatible"
    assert entity.dxf.insert.y == 2, "is not Vec3 compatible"
    assert entity.dxf.insert.z == 3, "is not Vec3 compatible"
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.insert == (0, 0, 0)


@pytest.mark.parametrize(
    "txt,ver",
    [
        (ENTITY_R2000, DXF2000),
        (ENTITY_R12, DXF12),
    ],
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    attdef = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    attdef.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    attdef.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


@pytest.mark.parametrize(
    "invalid_text",
    [
        "test\ntext\r",
        "test\r\ntext",
        "testtext^",
        "test\ntext^",
        "test\ntext^\r",
    ],
)
def test_removing_invalid_chars_at_setting_content(invalid_text):
    txt = Text()
    txt.dxf.text = invalid_text
    assert txt.dxf.text == "testtext"


@pytest.fixture
def text():
    return Text.new(handle="ABBA", owner="0")


def test_set_alignment(text):
    text.set_placement((2, 2), align=TextEntityAlignment.TOP_CENTER)
    assert text.dxf.halign == 1
    assert text.dxf.valign == 3
    assert text.dxf.align_point == (2, 2)


def test_set_fit_alignment(text):
    text.set_placement((2, 2), (4, 2), align=TextEntityAlignment.FIT)
    assert text.dxf.halign == 5
    assert text.dxf.valign == 0
    assert text.dxf.insert == (2, 2)
    assert text.dxf.align_point == (4, 2)


def test_reset_fit_alignment(text):
    text.set_placement((2, 2), (4, 2), align=TextEntityAlignment.FIT)
    text.set_placement((3, 3), (5, 3))
    assert text.dxf.halign == 5
    assert text.dxf.valign == 0
    assert text.dxf.insert == (3, 3)
    assert text.dxf.align_point == (5, 3)


def test_resetting_location_raises_value_error_for_missing_point(text):
    text.set_placement((2, 2), (4, 2), align=TextEntityAlignment.FIT)
    with pytest.raises(ValueError):
        text.set_placement((3, 3))


def test_get_align_enum(text):
    text.dxf.halign = 1
    text.dxf.valign = 3
    assert text.get_align_enum() == TextEntityAlignment.TOP_CENTER


def test_get_pos_enum_TOP_CENTER(text):
    text.set_placement((2, 2), align=TextEntityAlignment.TOP_CENTER)
    align, p1, p2 = text.get_placement()
    assert align == TextEntityAlignment.TOP_CENTER
    assert p1 == (2, 2)
    assert p2 is None


def test_get_pos_LEFT(text):
    text.set_placement((2, 2))
    align, p1, p2 = text.get_placement()
    assert align == TextEntityAlignment.LEFT
    assert p1 == (2, 2)
    assert p2 is None


def test_transform_interface():
    text = Text()
    text.dxf.insert = (1, 0, 0)
    text.transform(Matrix44.translate(1, 2, 3))
    assert text.dxf.insert == (2, 2, 3)

    # optimized translate
    text.dxf.align_point = (3, 2, 1)
    text.translate(1, 2, 3)
    assert text.dxf.insert == (3, 4, 6)
    assert text.dxf.align_point == (4, 4, 4)


def test_fit_length(text):
    text.set_placement((2, 2), (4, 2), align=TextEntityAlignment.FIT)
    assert text.fit_length() == 2

    # remove align point
    del text.dxf.align_point
    assert text.fit_length() == 0


def test_default_font_name(text):
    assert text.font_name() == "arial.ttf"


@pytest.fixture
def text2():
    return Text.new(
        dxfattribs={
            "text": "TEXT",
            "height": 1.0,
            "width": 1.0,
            "rotation": 0,
            "layer": "text",
        }
    ).set_placement((0, 0, 0), align=TextEntityAlignment.LEFT)


@pytest.mark.parametrize("rx, ry", [(1, 1), (-1, 1), (-1, -1), (1, -1)])
def test_scale_and_reflexion(rx, ry, text2):
    insert = Vec3(0, 0, 0)
    m = Matrix44.chain(
        Matrix44.scale(2 * rx, 3 * ry, 1),
        Matrix44.z_rotate(math.radians(45)),
        Matrix44.translate(3 * rx, 3 * ry, 0),
    )

    text2.transform(m)
    check_point = m.transform(insert)
    ocs = text2.ocs()
    assert ocs.to_wcs(text2.dxf.insert).isclose(check_point)
    assert math.isclose(text2.dxf.height, 3.0)
    assert math.isclose(text2.dxf.width, 2.0 / 3.0)


def test_non_uniform_scaling(text2):
    text2.rotate_z(math.radians(30))
    text2.scale(1, 2, 1)
    assert math.isclose(text2.dxf.oblique, 33.004491598883064)


def test_is_backward(text):
    assert text.is_backward is False


def test_set_backward(text):
    text.is_backward = True
    assert text.is_backward is True


def test_is_upside_down(text):
    assert text.is_upside_down is False


def test_set_is_upside_down(text):
    text.is_upside_down = True
    assert text.is_upside_down is True


def test_get_pos_handles_missing_align_point():
    """Any text alignment except LEFT requires and uses the align_point
    attribute as text location point. But there are real world example from
    AutoCAD which do not provide the align_point even it is required.

    In this case the get_pos() method returns the insert attribute.

    """
    text = Text()
    text.dxf.halign = 1  # center
    text.dxf.valign = 1  # bottom
    text.dxf.insert = (1, 2)
    text.dxf.align_point = (3, 4)  # the real alignment point

    # the expected and correct align point:
    alignment, p1, p2 = text.get_placement()
    assert p1 == (3, 4)
    assert p2 is None  # only used for FIT and ALIGNED

    # remove the align point
    del text.dxf.align_point

    alignment, p1, p2 = text.get_placement()
    assert p1 == (1, 2)  # use the insert point instead
    assert p2 is None  # only used for FIT and ALIGNED

MALFORMED_TEXT = """0
TEXT
5
0
6
LT_EZDXF
8
LY_EZDXF
330
0
100
AcDbEntity
10
1.0
20
2.0
30
3.0
40
1.0
1
TEXTCONTENT
50
0.0
51
0.0
7
STY_EZDXF
41
1.0
71
0
72
3
11
3.0
21
4.0
31
5.0
100
AcDbText
73
1
62
7
"""


def test_malformed_text():
    line = Text.from_text(MALFORMED_TEXT)
    assert line.dxf.layer == "LY_EZDXF"
    assert line.dxf.linetype == "LT_EZDXF"
    assert line.dxf.style == "STY_EZDXF"
    assert line.dxf.color == 7
    assert line.dxf.text == "TEXTCONTENT"
    assert line.dxf.insert.isclose((1, 2, 3))
    assert line.dxf.align_point.isclose((3, 4, 5))
    assert line.dxf.halign == 3
    assert line.dxf.valign == 1



# Copyright (c) 2010-2023 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence
from ezdxf.math import Vec3, Bezier4P, UVec

# These are low-level interpolation tools for B-splines, which can not be
# integrated into other curve related modules!
__all__ = ["cubic_bezier_interpolation", "tangents_cubic_bezier_interpolation"]


def cubic_bezier_interpolation(
    points: Iterable[UVec],
) -> Iterable[Bezier4P[Vec3]]:
    """Returns an interpolation curve for given data `points` as multiple cubic
    Bézier-curves. Returns n-1 cubic Bézier-curves for n given data points,
    curve i goes from point[i] to point[i+1].

    Args:
        points: data points

    """
    from ezdxf.math.linalg import tridiagonal_matrix_solver

    # Source: https://towardsdatascience.com/b%C3%A9zier-interpolation-8033e9a262c2
    pnts = Vec3.tuple(points)
    if len(pnts) < 3:
        return

    num = len(pnts) - 1

    # setup tri-diagonal matrix (a, b, c)
    b = [4.0] * num
    a = [1.0] * num
    c = [1.0] * num
    b[0] = 2.0
    b[num - 1] = 7.0
    a[num - 1] = 2.0

    # setup right-hand side quantities
    points_vector = [pnts[0] + 2.0 * pnts[1]]
    points_vector.extend(
        2.0 * (2.0 * pnts[i] + pnts[i + 1]) for i in range(1, num - 1)
    )
    points_vector.append(8.0 * pnts[num - 1] + pnts[num])

    # solve tri-diagonal linear equation system
    solution = tridiagonal_matrix_solver([a, b, c], points_vector)
    control_points_1 = Vec3.list(solution.rows())
    control_points_2 = [
        p * 2.0 - cp for p, cp in zip(pnts[1:], control_points_1[1:])
    ]
    control_points_2.append((control_points_1[num - 1] + pnts[num]) / 2.0)

    for defpoints in zip(
        pnts, control_points_1, control_points_2, pnts[1:]
    ):
        yield Bezier4P(defpoints)


def tangents_cubic_bezier_interpolation(
    fit_points: Sequence[Vec3], normalize=True
) -> list[Vec3]:
    if len(fit_points) < 3:
        raise ValueError("At least 3 points required")

    curves = list(cubic_bezier_interpolation(fit_points))
    tangents = [
        (curve.control_points[1] - curve.control_points[0]) for curve in curves
    ]

    last_points = curves[-1].control_points
    tangents.append(last_points[3] - last_points[2])
    if normalize:
        tangents = [t.normalize() for t in tangents]
    return tangents



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License

import pytest

import ezdxf
from ezdxf.entities.dimension import RadialDimensionLarge
from ezdxf.lldxf.const import DXF2010
from ezdxf.lldxf.tagwriter import TagCollector

ezdxf.options.preserve_proxy_graphics()


@pytest.fixture()
def dim():
    return RadialDimensionLarge.from_text(LARGE_RADIAL_DIM)


def test_load_dimension(dim):
    assert dim.dxf.defpoint == (46, 54, 0)
    assert dim.dxf.chord_point == (46, 54, 0)
    assert dim.dxf.override_center == (47, 53, 0)
    assert dim.dxf.jog_point == (49, 52, 0)
    assert dim.dxf.unknown2 == 0.0


def test_export_dimension(dim):
    tagwriter = TagCollector(dxfversion=DXF2010)
    dim.export_dxf(tagwriter)
    assert (10, 46) in tagwriter.tags
    assert (13, 46) in tagwriter.tags
    assert (14, 47) in tagwriter.tags
    assert (15, 49) in tagwriter.tags
    assert (40, 0.0) in tagwriter.tags


LARGE_RADIAL_DIM = """  0
LARGE_RADIAL_DIMENSION
5
37C
102
{ACAD_REACTORS
330
393
102
}
330
1F
100
AcDbEntity
8
0
100
AcDbDimension
280
0
2
*D15
10
46.0
20
54.0
30
0.0
11
48.0
21
53.0
31
0.0
70
169
71
5
42
3.089400503595107
73
0
74
0
75
0
3
Standard
100
AcDbRadialDimensionLarge
13
46.0
23
54.0
33
0.0
14
47.0
24
53.0
34
0.0
15
49.0
25
52.0
35
0.0
40
0.0
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2023 Manfred Moitzi
# License: MIT License
from typing import Optional
from ezdxf.enums import InsertUnits

# Documentation: https://ezdxf.mozman.at/docs/concepts/units.html#insunits

MSP_METRIC_UNITS_FACTORS = {
    # length in units / factor = length in meters
    # length in meters * factor = length in units
    "km": 0.001,
    "m": 1.0,
    "dm": 10.0,
    "cm": 100.0,
    "mm": 1000.0,
    "µm": 1000000.0,
    "yd": 1.093613298,
    "ft": 3.280839895,
    "in": 39.37007874,
    "mi": 0.00062137119,
}

IN = 1
FT = 2
MI = 3
MM = 4
CM = 5
M = 6
KM = 7
YD = 10
DM = 14

IMPERIAL_UNITS = {
    InsertUnits.Inches,
    InsertUnits.Feet,
    InsertUnits.Miles,
    InsertUnits.Microinches,
    InsertUnits.Mils,
    InsertUnits.Yards,
    InsertUnits.USSurveyFeet,
    InsertUnits.USSurveyInch,
    InsertUnits.USSurveyYard,
    InsertUnits.USSurveyMile,
}

# Conversion factor from meters to unit
# 1 meter is ... [unit]
METER_FACTOR = [
    None,  # 0 = Unitless - not supported
    39.37007874,  # 1 = Inches
    3.280839895,  # 2 = Feet
    0.00062137119,  # 3 = Miles
    1000.0,  # 4 = Millimeters
    100.0,  # 5 = Centimeters
    1.0,  # 6 = Meters
    0.001,  # 7 = Kilometers
    None,  # 8 = Microinches = 1e-6 in
    None,  # 9 = Mils = 0.001 in
    1.093613298,  # 10 = Yards
    10000000000.0,  # 11 = Angstroms = 1e-10m
    1000000000.0,  # 12 = Nanometers = 1e-9m
    1000000.0,  # 13 = Microns = 1e-6m
    10.0,  # 14 = Decimeters = 0.1m
    0.1,  # 15 = Decameters = 10m
    0.01,  # 16 = Hectometers = 100m
    0.000000001,  # 17 = Gigameters = 1e+9 m
    1.0 / 149597870700,  # 18 = Astronomical units = 149597870700m
    1.0 / 9.46e15,  # 19 = Light years = 9.46e15 m
    1.0 / 3.09e16,  # 20 = Parsecs =  3.09e16 m
    None,  # 21 = US Survey Feet
    None,  # 22 = US Survey Inch
    None,  # 23 = US Survey Yard
    None,  # 24 = US Survey Mile
]


class DrawingUnits:
    def __init__(self, base: float = 1.0, unit: str = "m"):
        self.base = float(base)
        self.unit = unit.lower()
        self._units = MSP_METRIC_UNITS_FACTORS
        self._msp_factor = base * self._units[self.unit]

    def factor(self, unit: str = "m") -> float:
        return self._msp_factor / self._units[unit.lower()]

    def __call__(self, unit: str) -> float:
        return self.factor(unit)


class PaperSpaceUnits:
    def __init__(self, msp=DrawingUnits(), unit: str = "mm", scale: float = 1):
        self.unit = unit.lower()
        self.scale = scale
        self._msp = msp
        self._psp = DrawingUnits(1, self.unit)

    def from_msp(self, value: float, unit: str):
        drawing_units = value * self._msp(unit.lower())
        return drawing_units / (self._msp(self.unit) * self.scale)

    def to_msp(self, value: float, unit: str):
        paper_space_units = value * self.scale * self._psp.factor(unit)
        model_space_units = paper_space_units * self._msp.factor(self.unit)
        return model_space_units


# Layout units are stored as enum in the associated BLOCK_RECORD: BlockRecord.dxf.units
# or as  optional XDATA for all DXF versions
# 1000: "ACAD"
# 1001: "DesignCenter Data" (optional)
# 1002: "{"
# 1070: Autodesk Design Center version number
# 1070: Insert units: like 'units'
# 1002: "}"

# The units of the modelspace block record is always 0, the real modelspace
# units and therefore the document units are stored as enum in the header var
# $INSUNITS

# units stored as enum in BlockRecord.dxf.units
# 0 = Unitless
# 1 = Inches
# 2 = Feet
# 3 = Miles
# 4 = Millimeters
# 5 = Centimeters
# 6 = Meters
# 7 = Kilometers
# 8 = Microinches = 1e-6 in
# 9 = Mils = 0.001 in
# 10 = Yards
# 11 = Angstroms = 1e-10m
# 12 = Nanometers = 1e-9m
# 13 = Microns = 1e-6m
# 14 = Decimeters = 0.1m
# 15 = Decameters = 10m
# 16 = Hectometers = 100m
# 17 = Gigameters = 1e+9 m
# 18 = Astronomical units = 149597870700m = 1.58125074e−5 ly =  4.84813681e−6 Parsec
# 19 = Light years = 9.46e15 m
# 20 = Parsecs =  3.09e16 m
# 21 = US Survey Feet
# 22 = US Survey Inch
# 23 = US Survey Yard
# 24 = US Survey Mile
_unit_spec = [
    None,
    "in",
    "ft",
    "mi",
    "mm",
    "cm",
    "m",
    "km",
    "µin",
    "mil",
    "yd",
    "Å",
    "nm",
    "µm",
    "dm",
    "dam",
    "hm",
    "gm",
    "au",
    "ly",
    "pc",
    None,
    None,
    None,
    None,
]


def decode(enum: int) -> Optional[str]:
    return _unit_spec[int(enum)]


def conversion_factor(
    source_units: InsertUnits, target_units: InsertUnits
) -> float:
    """Returns the conversion factor to represent `source_units` in
    `target_units`.

    E.g. millimeter in centimeter :code:`conversion_factor(MM, CM)` returns 0.1,
    because 1 mm = 0.1 cm

    """
    try:
        source_factor = METER_FACTOR[source_units]
        target_factor = METER_FACTOR[target_units]
        if source_factor is None or target_factor is None:
            raise TypeError("Unsupported conversion.")
        return target_factor / source_factor

    except IndexError:
        raise ValueError("Invalid unit enum.")


def unit_name(enum: int) -> str:
    """Returns the name of the unit enum."""
    try:
        return InsertUnits(enum).name
    except ValueError:
        return "unitless"


ANGLE_UNITS = {
    0: "Decimal Degrees",
    1: "Degrees/Minutes/Seconds",
    2: "Grad",
    3: "Radians",
}


def angle_unit_name(enum: int) -> str:
    """Returns the name of the angle unit enum."""
    return ANGLE_UNITS.get(enum, f"unknown unit <{enum}>")
